static char *CopyrightIdentifier(void) { return "@(#)exansi9t.cc 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved."; }

#include <stdlib.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include <unistd.h>	// for read(2)
#include <fcntl.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>

#include <string.h>
#include <ctype.h>

#include "GetOpt.h"
#include "MainCmd.h"

#include "ansi9t.h"

static const char *extractSansSpaces(const char *from,unsigned length)
{
	static char buffer[80];
	char *to = buffer;
	while (*from && length--) {
		if (isspace(*from)) ++from;
		else *to++=*from++;
	}
	*to=0;
	return buffer;
}

int
main(int argc,char **argv)
{
	char *devicename;
	int list=0;
	int verbose=0;
	int unique=0;

	GetOpt getopt(argc,argv,"luv");
	int c;
	int err=0;
	while ((c=getopt()) != -1 && !err) {
		switch (c) {
			case 'l':
				list=1;
				break;
			case 'u':
				unique=1;
				break;
			case 'v':
				verbose=1;
				break;
			case '?':
			default:
				err=1;
		}
	}

	if (getopt.optind+1 == argc) {
		devicename=argv[getopt.optind];
	}
	else {
		cerr << "Must specify one devicename\n";
		err=1;
	}

	if (err) {
		cerr << "Usage: " << argv[0] << " -l -u -v devicename\n"
		     << flush;
		return 1;
	}

	static char buf[65535];
	int fd;

	if ((fd=open(devicename,O_RDONLY)) == -1) {
		cerr << "Can't open " << devicename << "\n" << flush;
		exit(1);
	}

	ofstream outstream;
	int filecount=0;
	char *filename=0;
	int markcount=0;
	int lng=0;
	while (markcount < 2) {
		int num=read(fd,buf,64512);
		switch (num) {
			case -1:
				cerr << "Read Error\n" << flush;
				exit(1);
			case 0:
				if (lng) {
					if (verbose)
						cerr << "Wrote " << lng
						     << "bytes\n" 
						     << flush;
					outstream.close();
					delete[] filename;
					filename=0;
					++filecount;
					lng=0;
				}
				if (verbose) cerr << "Tape Mark\n" << flush;
				++markcount;
				break;
			case 80:
				markcount=0;
				if (!lng) {	// First record, test for label
					union AnsiHdr *u = (union AnsiHdr *)buf;
					if (!strncmp(u->c.labelid,"VOL",3)
					 || !strncmp(u->c.labelid,"EOF",3)
					 || !strncmp(u->c.labelid,"EOV",3)
					 || !strncmp(u->c.labelid,"UTL",3)
					 || !strncmp(u->c.labelid,"UVL",3)) {
						if (verbose) {
							cerr << extractSansSpaces(u->c.labelid,3);
							cerr << extractSansSpaces(u->c.labelno,1) << ":\n" << flush;
						}	
						if (filename) delete[] filename;
						filename=0;
						break;
					}
					else if (!strncmp(u->c.labelid,"UHL",3)) {
						if (verbose) {
							cerr << extractSansSpaces(u->c.labelid,3);
							cerr << extractSansSpaces(u->c.labelno,1) << ":\n" << flush;
						}	
						// ignore
						break;
					}
					else if (!strncmp(u->c.labelid,"HDR",3)) {
						if (verbose) {
							cerr << extractSansSpaces(u->c.labelid,3);
							cerr << extractSansSpaces(u->c.labelno,1) << ":\n" << flush;
						}	
						if (!strncmp(u->c.labelno,"1",1)) {
							ostrstream ost;
							if (!unique) {
								ost << "file"
								    << setfill('0')
								    << setw(4)
								    << atoi(extractSansSpaces(u->hdr1.filesequencenumber,4))
								    << ".";
							}
							ost << extractSansSpaces(u->hdr1.fileid,17) << ends;
							filename=ost.str();
						}
						// else ignore
						break;
					}
					// unrecognized label, fall through ...
				}
				// assume middle of file, fall through ...
			default:
				if (verbose)
					cerr << "Read " << num
					     << "bytes\n" << flush;
				markcount=0;
				if (!lng) {
					if (filename) {
						outstream.open(filename);
						if (!outstream.good()) {
							cerr << "Open failed of " 
							     << filename << "\n" 
							     << flush;
							exit(1);
						}
						if (list || verbose)
							cerr << filename << "\n" << flush;
					}
					else {
						cerr << "No ANSI HDR1 label before data file\n" << flush;
						exit(1);
					}
				}
				outstream.write(buf,num);
				lng+=num;
				break;
		}
	}
	if (verbose) cerr << "Two Tape Marks == End Of Media\n" << flush;
	exit(0);
}
