static char *CopyrightIdentifier(void) { return "@(#)dc3tape.cc 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved."; }

#define USEDIRECTORYACCESS

#include <stdlib.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include <unistd.h>	// for read(2)
#include <fcntl.h>
#include <errno.h>
extern int sys_nerr;
extern char *sys_errlist[];

#include <sys/types.h>
#include <sys/mtio.h>
#include <sys/ioctl.h>

#ifdef NEEDMISSINGPROTOTYPES
extern "C" {
int ioctl(int fd, int request, caddr_t arg);
}
#endif /* NEEDMISSINGPROTOTYPES */

#include <ctype.h>

#include "BaseType.h"
#include "GetOpt.h"
#include "MainCmd.h"

//#define zeropad(i,l)	dec << setfill('0') << setw(l) << i << setfill(' ')

class VolumeHeader {		// 512 bytes
private:
	char Identifier[12];
	char Partitions[14];
	char reserved1[2];
	Uint32_L FixedBlockLength;
	char reserved2[480];
public:
	VolumeHeader(unsigned long blocklength)
		{
			strncpy(Identifier,"DICOMVOLHDR",sizeof(Identifier));
			strncpy(Partitions,"ONEPARTITION",sizeof(Partitions));
			memset(reserved1,0,sizeof(reserved1));
			FixedBlockLength=blocklength;
			memset(reserved2,0,sizeof(reserved2));
		}

	void dump(ostream &ostr) const
		{
			ostr << "Volume Header:\n";
			ostr << "\tIdentifier:\t\t\t" << Identifier << "\n";
			ostr << "\tPartitions:\t\t\t" << Partitions << "\n";
			ostr << "\tFixed Block Length:\t\t"		<< FixedBlockLength << "\n";
			ostr << flush;
		}
};

class FileNumberEntry {		// 128 bytes
private:
	Uint32_L FileNumber;
	Uint32_L FileLengthInBytes;
	char FileID[72];
	char FileType[9];
	char reserved1[39];
public:
	FileNumberEntry(void)	{}

	FileNumberEntry(
			unsigned long filenumber,
			unsigned long filelengthinbytes,
			const char *fileid,
			const char *filetype)
		{
			FileNumber=filenumber;
			FileLengthInBytes=filelengthinbytes;
			strncpy(FileID,fileid,sizeof(FileID));
			strncpy(FileType,filetype,sizeof(FileType));
			memset(reserved1,0,sizeof(reserved1));
		}

	void dump(ostream &ostr) const
		{
			ostr << "\t# " << setw(6) << dec << FileNumber;
			ostr << "\t"   << setw(8) << FileLengthInBytes << " bytes";
			ostr << "\t("  << FileType << ")  ";
			ostr << "\t" << FileID;
			ostr << setfill(' ') << "\n";
			ostr << flush;
		}

	const char *getFileID(void) const	{ return FileID; }
	Uint32_L    getFileNumber(void) const	{ return FileNumber; }
	Uint32_L    getFileLength(void) const	{ return FileLengthInBytes; }
};

class DataFileHeader {		// 512 bytes
private:
	char Identifier[16];
	FileNumberEntry fne;	// 128 bytes
	char reserved1[368];
public:
	DataFileHeader(
			unsigned long filenumber,
			unsigned long filelengthinbytes,
			const char *fileid,
			const char *filetype)
		{
			strncpy(Identifier,"DICOMFILEHDR",sizeof(Identifier));
			fne=FileNumberEntry(filenumber,filelengthinbytes,fileid,filetype);
			memset(reserved1,0,sizeof(reserved1));
		}


	void dump(ostream &ostr) const
		{
			ostr << "Data File Header:\n";
			ostr << "\tIdentifier:\t" << Identifier << "\n";
			fne.dump(ostr);
			ostr << flush;
		}

	const char *getFileID(void) const	{ return fne.getFileID(); }
	Uint32_L    getFileLength(void) const	{ return fne.getFileLength(); }
};

class LogicalFileSystemDirectoryFixed {		// 512 bytes
public:
	char Identifier[14];
	char UseField[7];
	char reserved1[3];
	Uint32_L TotalNumberOfFiles;
	Uint32_L TotalNumberOfDicomFiles;
	Uint32_L DICOMDIRFileNumber;
	Uint32_L TotalBytesInFiles;
	char reserved2[472];

	LogicalFileSystemDirectoryFixed(void)
		{
			strncpy(Identifier,"DICOMMEDIADIR",sizeof(Identifier));
			strncpy(UseField,"UNUSED",sizeof(UseField));
			memset(reserved1,0,sizeof(reserved1));
			TotalNumberOfFiles=0;
			TotalNumberOfDicomFiles=0;
			DICOMDIRFileNumber=0;
			TotalBytesInFiles=0;
		}

	void dump(ostream &ostr,const char *title) const
		{
			ostr << "LogicalFileSystemDirectory (" << (title ? title : "") << "):\n";
			ostr << "\tIdentifier:\t\t\t"			<< Identifier << "\n";
			ostr << "\tUse Field:\t\t\t"			<< UseField << "\n";
			ostr << "\tTotal Number Of Files:\t\t"		<< TotalNumberOfFiles << "\n";
			ostr << "\tTotal Number Of Dicom Files:\t"	<< TotalNumberOfDicomFiles << "\n";
			ostr << "\tDICOMDIR File Number:\t\t"		<< DICOMDIRFileNumber << "\n";
			ostr << "\tTotal Bytes In Files:\t\t"		<< TotalBytesInFiles << "\n";
			ostr << flush;
		}
};

class LogicalFileSystemDirectory {
	LogicalFileSystemDirectoryFixed Fixed;
	FileNumberEntry *Entries;
	unsigned long NEntries;
public:
	LogicalFileSystemDirectory(
			unsigned long nfiles)
		{
			Entries=nfiles ? new FileNumberEntry[nfiles] : 0;
			NEntries=nfiles;
		}

	const char *getFixed(void)		{ return (const char *)&Fixed; }
	const char *getEntries(void)		{ return (const char *)Entries; }
	unsigned long getNumberOfEntries(void)	{ return NEntries; }

	void setInUse(void)	{ strncpy(Fixed.UseField,"INUSE",sizeof(Fixed.UseField)); }

	int addFileNumberEntry(
			unsigned long filenumber,
			unsigned long filelengthinbytes,
			const char *fileid,
			const char *filetype)
		{
			FileNumberEntry fne(filenumber,filelengthinbytes,fileid,filetype);
			if (filenumber <= NEntries) {
				Entries[filenumber-1]=fne;
				Fixed.TotalNumberOfFiles=Fixed.TotalNumberOfFiles+1;
				if (strcmp(filetype,"DICOM") == 0)
					Fixed.TotalNumberOfDicomFiles=Fixed.TotalNumberOfDicomFiles+1;
				else if (strcmp(filetype,"DICOMDIR") == 0)
					Fixed.DICOMDIRFileNumber=filenumber;
				Fixed.TotalBytesInFiles=Fixed.TotalBytesInFiles+filelengthinbytes;
			}
			else {
				cerr << "LogicalFileSystemDirectory::addFileNumberEntry()"
				     << " error - filenumber " << filenumber
				     << " > number of files " << NEntries
				     << "\n" << flush;
				return -1;
			}
			return 0;
		}
};

struct FileNameAndNumber {
	daddr_t number;
	char *name;
	unsigned long length;
};

static int fileNameAndNumberCompare(const void *i,const void *j)
{
	return ((FileNameAndNumber*)i)->number - ((FileNameAndNumber*)j)->number;
}

class SequentialMediaFileSystem {
	const char *Device;
	size_t BlockSize;
	size_t LastBlockModulo;
	char *Buffer;
	int fd;

	int open(int mode)
		{
			if ((fd=::open(Device,mode)) == -1) {
				cerr << "Can't open <" << Device << ">\n" << flush;
			}
			return fd;
		}

	size_t sizeModulo(size_t nwanted)
		{
			return LastBlockModulo ? ((nwanted-1)/LastBlockModulo+1)*LastBlockModulo : nwanted;
		}

	size_t sizeModuloAndClear(size_t nwanted)
		{
			//cerr << "write: want to write " << nwanted << " bytes\n" << flush;

			size_t nwrite = sizeModulo(nwanted);
			if (nwrite-nwanted > 0)
				memset(Buffer+nwanted,0,nwrite-nwanted);

			//cerr << "write: have to write " << nwrite << " bytes\n" << flush;

			return nwrite;
		}

	int write(size_t nwanted)
		{
			size_t nwrite = sizeModuloAndClear(nwanted);
			size_t nwritten;
			if ((nwritten=::write(fd,Buffer,nwrite)) == -1) {
				cerr << "Can't write to <" << Device << ">"
				     << " perror=" << errno
				     << " " << (errno<sys_nerr ? sys_errlist[errno] : "") << " "
				     << "\n" << flush;
				return -1;
			}
			if (nwritten != nwrite) {
				cerr << "Can't write to <" << Device << ">"
				     << " would only write " << nwritten
				     << " not " << nwrite << " bytes\n" << flush;
				return -1;
			}
			return nwritten;
		}

	int writeHeader(const char *ptr,size_t size)
		{
			memcpy(Buffer,ptr,size);
			if (write(size) == -1
			 || writeTapeMark() == -1)
				return -1;
			else
				return 0;
		}

	int writeVolumeHeader(VolumeHeader& volhdr)
		{
			return writeHeader((const char *)&volhdr,sizeof(VolumeHeader));
		}

	int writeDataFileHeader(DataFileHeader& dhdr)
		{
			return writeHeader((const char *)&dhdr,sizeof(DataFileHeader));
		}

	int writeFile(const char*filename,unsigned long &bytecount,
			ostream& log,int verbose)
		{
			(void)log;
			(void)verbose;
			ifstream infile(filename);
			if (!infile.rdbuf()->is_open()) {
				cerr << "Can't open file <" << filename
				     << ">\n" << flush;
				return -1;
			}
			bytecount=0;
			size_t nread;
			do {
				infile.read(Buffer,BlockSize);
				if ((nread=infile.gcount())) {
					bytecount+=nread;
					if (write(nread) == -1) return -1;
				}
			} while (nread && infile.good());

			if (!bytecount || writeTapeMark() == -1) return -1;
		}

	int writeLogicalFileSystemDirectory(LogicalFileSystemDirectory& lfsd)
		{
			size_t 		fixedsize = sizeof(LogicalFileSystemDirectoryFixed);
			memcpy(Buffer,lfsd.getFixed(),fixedsize);
			size_t		bufcnt	  = fixedsize;
			long		size      = sizeof(FileNumberEntry)*lfsd.getNumberOfEntries();
			const char *	ptr	  = lfsd.getEntries();

			do {
				//cerr << "size " << size << "\n" << flush;
				//cerr << "bufcnt " << bufcnt << "\n" << flush;
				size_t bufleft=BlockSize-bufcnt;
				//cerr << "bufleft " << bufleft << "\n" << flush;
				size_t bufadd=(size_t)((size > bufleft) ? bufleft : size);
				//cerr << "bufadd " << bufadd << "\n" << flush;
				if (bufadd) {
					//cerr << "adding\n" << flush;
					memcpy(Buffer+bufcnt,ptr,bufadd);
					ptr+=bufadd;
					bufcnt+=bufadd;
					//cerr << "added\n" << flush;
					//cerr << "bufcnt now " << bufcnt << "\n" << flush;
				}
				if (write(bufcnt) == -1) return -1;
				size-=bufadd;
				//cerr << "size now " << size << "\n" << flush;
				bufcnt=0;
			} while (size > 0);
			if (writeTapeMark() == -1) return -1;
			return 0;
		}

	int read(size_t nwanted)
		{
			size_t nread;
			if ((nread=::read(fd,Buffer,nwanted)) == -1) {
				cerr << "Can't read from <" << Device << ">"
				     << " perror=" << errno
				     << " " << (errno<sys_nerr ? sys_errlist[errno] : "") << " "
				     << "\n" << flush;
			}
			return nread;
		}

	int command(short op,daddr_t count,daddr_t& resid,daddr_t& fileno,daddr_t& blkno)
		{
          		struct mtop mt_command;
          		struct mtget mt_status;
			mt_command.mt_op = op;
			mt_command.mt_count = count;
			if (ioctl(fd, MTIOCTOP, (caddr_t)&mt_command) == -1) {
				cerr << "ioctl command fail"
				     << " op=" << op << " count=" << count
				     << " perror=" << errno
				     << " " << (errno<sys_nerr ? sys_errlist[errno] : "") << " "
				     << "\n" << flush;
				return -1;
			}
			if (ioctl(fd, MTIOCGET, (caddr_t)&mt_status) == -1) {
				cerr << "ioctl status fail"
				     << " perror=" << errno
				     << " " << (errno<sys_nerr ? sys_errlist[errno] : "") << " "
				     << "\n" << flush;
				return -1;
			}
			//cerr << "mt_status.mt_resid " << mt_status.mt_resid << "\n" << flush;
			//cerr << "mt_status.mt_fileno " << mt_status.mt_fileno << "\n" << flush;
			//cerr << "mt_status.mt_blkno " << mt_status.mt_blkno << "\n" << flush;
			resid=mt_status.mt_resid;
			fileno=mt_status.mt_fileno;
			blkno=mt_status.mt_blkno;
			if (mt_status.mt_erreg) {
				cerr << "mt_status error"
				     << " mt_dsreg=" << mt_status.mt_dsreg
				     << " mt_erreg=" << mt_status.mt_erreg
				     << "\n" << flush;
				return -1;
			}
			return 0;
		}

	int command(short op,daddr_t count)
		{
			daddr_t resid;
			daddr_t fileno;
			daddr_t blkno;
			return command(op,count,resid,fileno,blkno);
		}

	int status(daddr_t& resid,daddr_t& fileno,daddr_t& blkno)
		{
			//cerr << "status:\n" << flush;
			return command(MTNOP,1,resid,fileno,blkno);
		}

	int rewind(void)
		{
			//cerr << "rewind:\n" << flush;
			return command(MTREW,1);
		}

	int eom(void)
		{
			//cerr << "eom:\n" << flush;
			return command(MTEOM,1);
		}

	int fsf(daddr_t count)
		{
			//cerr << "fsf: " << count << "\n" << flush;
			return command(MTFSF,count);
		}

	int writeTapeMark(void)
		{
			//cerr << "writeTapeMark:\n" << flush;
			return command(MTWEOF,1);
		}

	void zeroBuffer(unsigned long n)
		{
			memset(Buffer,0,(int)n);
		}
public:
	SequentialMediaFileSystem(
			const char *device,
			size_t blocksize,
			size_t lastblockmodulo)
		{
			Device=device;
			BlockSize=blocksize;
			LastBlockModulo=lastblockmodulo;
			Buffer=new char[BlockSize];
		}

	~SequentialMediaFileSystem()
		{
			if (Buffer) delete[] Buffer;
		}

	int Create(
			unsigned long nfiles,
			const char *const *filenames,
			const char *dicomdir,
			ostream& log,int verbose,
			int dontusedirectory)
		{

			if (open(O_RDWR) == -1) return -1;

			if (rewind() == -1) return -1;

			// First file - VolumeHeader

			VolumeHeader volhdr(BlockSize);
			if (writeVolumeHeader(volhdr) == -1) return -1;

			// Second File - Logical File System Directory (empty and unused)

			LogicalFileSystemDirectory emptylfsd(0);
			if (writeLogicalFileSystemDirectory(emptylfsd) == -1) return -1;

			// NB. number of files includes the two DICOMDIRs */

			LogicalFileSystemDirectory lfsd(nfiles+2);

			// Third & Fourth Files - DICOMDIR preceded by its header

			{
				DataFileHeader dhdr(1,0,dicomdir,"DICOMDIR");
				if (writeDataFileHeader(dhdr) == -1) return -1;
				unsigned long bytecount;
				if (writeFile(dicomdir,bytecount,log,verbose) == -1) return -1;
				lfsd.addFileNumberEntry(1,bytecount,dicomdir,"DICOMDIR");
			}

			// Data Files

			size_t n=0;
			while (n < nfiles) {
				int filenumber=n+2;
				if (verbose)
					log << "Doing file # " << filenumber << "\n" << flush;

				DataFileHeader dhdr(filenumber,0,filenames[n],"DICOM");
				if (writeDataFileHeader(dhdr) == -1) return -1;
				unsigned long bytecount;
				if (writeFile(filenames[n],bytecount,log,verbose) == -1) {
					cerr << "Can't read file <" << filenames[n]
					     << "> - doesn't exit or is empty\n" << flush;
					return -1;
				}

				lfsd.addFileNumberEntry(filenumber,bytecount,filenames[n],"DICOM");
				++n;
			}

			// Third and Second Last Files - DICOMDIR preceded by its header

			{
				DataFileHeader dhdr(nfiles+2,0,dicomdir,"DICOMDIR");
				if (writeDataFileHeader(dhdr) == -1) return -1;
				unsigned long bytecount;
				if (writeFile(dicomdir,bytecount,log,verbose) == -1) return -1;
				lfsd.addFileNumberEntry(nfiles+2,bytecount,dicomdir,"DICOMDIR");
			}

			// Last File - filled and used Logical File System Directory

			lfsd.setInUse();

			if (writeLogicalFileSystemDirectory(dontusedirectory ? emptylfsd : lfsd) == -1) return -1;

			if (writeTapeMark() == -1) return -1;

			if (::close(fd) == -1) return -1;

			return 0;
		}

	int Update(
			int nfiles,
			const char *const *filenames,
			const char *dicomdir,
			ostream& log,int verbose,
			int dontusedirectory)
		{
			(void)nfiles;
			(void)filenames;
			(void)dicomdir;
			(void)log;
			(void)verbose;
			(void)dontusedirectory;
			return 0;
		}

#ifdef USEDIRECTORYACCESS

	int ReadOrList(
			unsigned long nfilestoget,
			const char *const *filenamestoget,
			const char *dicomdirtoget,
			ostream& listout,ostream& log,
			int verbose,int readfiles,int listfiles,
			int dontusedirectory)
		{
			if (open(O_RDONLY) == -1) return -1;

			if (rewind() == -1) return -1;
			int num=read(64512);
			if (num == -1 || num == 0 || strcmp(Buffer,"DICOMVOLHDR") != 0) {
				cerr << "DICOM Volume Header not present\n" << flush;
				return -1;
			}
			VolumeHeader *vhdr = (VolumeHeader *)Buffer;
			if (verbose) vhdr->dump(log);

			LogicalFileSystemDirectoryFixed *lfsdf =
				(LogicalFileSystemDirectoryFixed *)Buffer;

			if (!dontusedirectory) {
				if (fsf(1) == -1) return -1;

				num=read(64512);
				if (num == -1 || num == 0 || strcmp(Buffer,"DICOMMEDIADIR") != 0) {
					cerr << "Leading Logical File System Directory not present\n" << flush;
					return -1;
				}


				if (strcmp(lfsdf->UseField,"INUSE") != 0) {
					if (verbose) lfsdf->dump(log,"Inactive Leading");

					// Get the LFSD from the end of the tape
					// assuming the device can't position to absolute
					// file numbers and can't backspace ...

					if (rewind() == -1) return -1;	// In case confused
					if (eom() == -1) return -1;
					if (fsf(0) == -1) return -1;	// Need to do something useless to set status :(
					daddr_t resid;
					daddr_t fileno;
					daddr_t blkno;
					if (status(resid,fileno,blkno) == -1) return -1;
					if (rewind() == -1) return -1;
					if (fsf(fileno-1) == -1) return -1;
					num=read(64512);
					if (num == -1 || num == 0 || strcmp(Buffer,"DICOMMEDIADIR") != 0) {
						cerr << "Trailing Logical File System Directory not present\n" << flush;
						return -1;
					}
				}
				if (verbose) lfsdf->dump(log,"Active");
			}

			// create an array of wanted files
			// the last entry of which is the DICOMDIR
			// fill in now if specific request
			// else leave empty and create later with entire directory

			FileNameAndNumber *files=0;
			unsigned long nfilesalready=0;
			if (nfilestoget || dicomdirtoget) {
				nfilesalready=nfilestoget;
				if (dicomdirtoget) ++nfilesalready;
				files=new FileNameAndNumber[nfilesalready];
				int i=0;
				while (i<nfilestoget) {
					files[i].number=-1;
					files[i].length=0;
					files[i].name=new char[strlen(filenamestoget[i])+1];
					strcpy(files[i].name,filenamestoget[i]);
					++i;
				}
				if (dicomdirtoget) {
					files[nfilestoget].number=-1;
					files[nfilestoget].length=0;
					files[nfilestoget].name=new char[strlen(dicomdirtoget)+1];
					strcpy(files[nfilestoget].name,dicomdirtoget);
				}
			}

			// zero is unknown ... never valid because always DICOMDIR times 2 at least ...

			unsigned long nfilesontape=0;
			if (!dontusedirectory) nfilesontape=lfsdf->TotalNumberOfFiles;

			if (!nfilesalready && nfilesontape) {
				// less 1 because we don't want both DICOMDIR copies !
				files=new FileNameAndNumber[nfilesontape-1];
				int i=0;
				while (i<nfilesontape-1) {
					files[i].number=-1;
					files[i].length=0;
					files[i].name=0;
					++i;
				}
			}

			// at this point we either have:
			// - files filled with wanted names and unknown places 
			// - files filled with unknown names & places (want all)
			// - files == 0 (not using directory & want all)

			// use the directory to fill in the places if in use ...

			if (files && !dontusedirectory && strcmp(lfsdf->UseField,"INUSE") == 0) {
				size_t fixedsize=sizeof(LogicalFileSystemDirectoryFixed);
				const char *ptr=Buffer+fixedsize;
				size_t bytesleft=num-fixedsize;
				unsigned long nfiles=nfilesontape;
				while (nfiles--) {
					char fnebuf[sizeof(FileNumberEntry)];
					size_t needbytes=sizeof(FileNumberEntry);
					while (needbytes > 0) {
						//cerr << "bytesleft " << bytesleft << "\n" << flush;
						//cerr << "needbytes " << needbytes << "\n" << flush;
						if (bytesleft <= 0) {
							//cerr << "bytesleft " << bytesleft << "\n" << flush;
							if ((num=read(64512)) > 0) {
								ptr=Buffer;
								bytesleft=num;
							}
							else {
								cerr << "Premature end of LFSD\n"
								     << flush;
								return -1;
							}
						}
						size_t addbytes=(needbytes < bytesleft)
							? needbytes
							: bytesleft;
						//cerr << "addbytes " << addbytes << "\n" << flush;
						memcpy(fnebuf,ptr,addbytes);
						ptr+=addbytes;
						bytesleft-=addbytes;
						needbytes-=addbytes;
					}
					FileNumberEntry *fne=(FileNumberEntry *)fnebuf;
					if (listfiles) fne->dump(listout);
					if (nfilesalready) {
						int i=0;
						while (i<nfilesalready) {
							if (files[i].name && strcmp(fne->getFileID(),files[i].name) == 0) {
								files[i].number=fne->getFileNumber();
								files[i].length=fne->getFileLength();
								break;
							}
							++i;
						}
					}
					else {
						unsigned long i;
						const char *s;
						if (strcmp(s=fne->getFileID(),"DICOMDIR") == 0) {
							i=nfilesontape-2;
						}
						else {
							i=fne->getFileNumber()-2;
						}
						files[i].name=new char[strlen(s)+1];
						strcpy(files[i].name,s);
						files[i].number=fne->getFileNumber();
						files[i].length=fne->getFileLength();
					}
				}
				if (!nfilesalready) nfilesalready=nfilesontape-1;
			}

			// At this point nfilesalready is the size of files[] 
			// for all scenarios.

			if (readfiles) {

				// sort to allow reading in sequence ...
				if (nfilesalready) qsort((void *)files,(int)nfilesalready,
					sizeof(FileNameAndNumber),
					fileNameAndNumberCompare);

				if (rewind() == -1) return -1;
				if (fsf(4) == -1) return -1;
				long tapefileno=4;	// First data file header (skip leading DICOMDIR)

				int i=0;
				while (1) {
					if (files && files[i].number > 0) {
						cerr << "#" << dec << setw(6) << files[i].number
						     << " " << files[i].name
						     << " length=" << files[i].length
						     << "\n" << flush;
					}
					long wantfileno=(files && files[i].number > 0)
						? ((files[i].number-1)*2+2)	// we know where to go
						: tapefileno;			// we don't ... just get the next one
					long skipcount=wantfileno-tapefileno;
					cerr << "at=" << tapefileno
					     << " want=" << wantfileno
					     << " skip=" << skipcount
					     << "\n" << flush;
					if (skipcount > 0 && fsf(skipcount) == -1) return -1;
					tapefileno=wantfileno;
					int num=read(64512);
					if (num == -1 || num == 0) {
						cerr << "Premature End of Media\n" << flush;
						return -1;
					}
					else if (strcmp(Buffer,"DICOMMEDIADIR") == 0
					      && (!files || files[i].number <= 0)
					) {
						break;	// Trailing LFSD marks end of data files
					}
					else if (strcmp(Buffer,"DICOMFILEHDR") != 0
					 || (num != sizeof(DataFileHeader)
					  && num != sizeModulo(sizeof(DataFileHeader)))
					) {
						cerr << "DICOM File Header not present\n" << flush;
						return -1;
					}
					DataFileHeader *dhdr = (DataFileHeader *)Buffer;
					if (verbose) dhdr->dump(log);

					char *usefileid=new char[strlen(dhdr->getFileID())+1];
					strcpy(usefileid,dhdr->getFileID());

					int wantit;
					if (files) {
						if (files[i].number > 0) {
							if (strcmp(usefileid,files[i].name) != 0) {
								cerr << "Expected " << files[i].name
								     << "but got " << usefileid
								     << "\n" << flush;
								return -1;
							}
							wantit=1;
						}
						else {
							// check all files[0..nfilesalready].name to see if wanted
							wantit=0;
							int j=0;
							while (j<nfilesalready) {
								if (strcmp(usefileid,files[j++].name) == 0) {
									wantit=1;
									break;
								}
							}
						}
					}
					else {
						wantit=1;
					}

					unsigned long wantbytes=dhdr->getFileLength();
					if (files && wantbytes == 0) wantbytes=files[i].length;

					if (wantbytes == 0 && LastBlockModulo != 0) {
						cerr << "Warning - <" << usefileid
						     << "> file length unknown and may be padded"
						     << "\n" << flush;
					}

					if (fsf(1) == -1) return -1;	// We didn't read past the tape mark yet
					++tapefileno;

					// Get the data file ...

					if (wantit) {
						ofstream useoutfile(usefileid);
						if (!useoutfile.rdbuf()->is_open()) {
							cerr << "Can't open file <" << usefileid
							     << ">\n" << flush;
							return -1;
						}

						while (1) {
							num=read(64512);
							cerr << "Read num=" << num << " wantbytes=" << wantbytes << "\n" << flush;
							if (num == 0) {
								++tapefileno;	// Have read past tape mark
								if (wantbytes > 0) {
									cerr << "Warning - Premature end of data <" << usefileid
									     << ">\n" << flush;
								}
								break;
							}
							else if (num == -1) {
								cerr << "Can't read from tape\n" << flush;
								return -1;
							}
							useoutfile.write(Buffer,(unsigned int)(wantbytes && wantbytes < num ? wantbytes : num));
							if (!useoutfile.good()) {
								cerr << "Can't write to file <" << files[i].name
						  		   << ">\n" << flush;
								return -1;
							}
							if (wantbytes) wantbytes-=num;
						} 
						useoutfile.close();
						if (usefileid) delete[] usefileid;
					}
					else {
						// Don't want it ... so blow past it
						if (fsf(1) == -1) return -1;
						++tapefileno;
					}

					if (files && files[i].number > 0) {
						++i;
						if (i >= nfilesalready) break;
					}
				}
			}

			if (::close(fd) == -1) return -1;

			return 0;
		}

#else // USEDIRECTORYACCESS

	int List(
			const char *dicomdir,
			ostream& dirout,ostream& log,int verbose)
		{
			if (open(O_RDONLY) == -1) return -1;

			// Get the LFSD from the end of the tape
			// assuming the device can't position to absolute
			// file numbers and can't backspace ...

			if (rewind() == -1) return -1;
			if (eom() == -1) return -1;
			if (fsf(0) == -1) return -1;	// Need to do something useless to set status :(
			daddr_t resid;
			daddr_t fileno;
			daddr_t blkno;
			if (status(resid,fileno,blkno) == -1) return -1;
			if (rewind() == -1) return -1;
			if (fsf(fileno-1) == -1) return -1;

			enum {
				ExpectVolumeHeader,
				ExpectLFSD,
				ExpectDataFileHeader,
				ExpectDataFile,
				ExpectDataFileHeaderOrLFSD,
				ExpectDataFileHeaderOrEnd
			} expect = ExpectLFSD;
			int markcount=0;
			unsigned long bytecount=0;
			while (markcount < 2) {
				int num=read(64512);
				switch (num) {
					case -1:
						return -1;
					case 0:
						//if (verbose) log << "Tape Mark\n" << flush;
						++markcount;
						bytecount=0;
						break;
					default:
						markcount=0;
						if      (bytecount == 0
						      && strcmp(Buffer,"DICOMVOLHDR") == 0
						      && (num == sizeof(VolumeHeader)
						       || num == sizeModulo(sizeof(VolumeHeader)))
						) {
							if (expect != ExpectVolumeHeader) {
								cerr << "Unexpected Volume Header\n" << flush;
								return -1;
							}
							VolumeHeader *volhdr = (VolumeHeader *)Buffer;
							if (verbose) volhdr->dump(dirout);
							expect=ExpectLFSD;
						}
						else if (bytecount == 0
						      && strcmp(Buffer,"DICOMFILEHDR") == 0
						      && (num == sizeof(DataFileHeader)
						       || num == sizeModulo(sizeof(DataFileHeader)))
						) {
							if (expect != ExpectDataFileHeader
							 && expect != ExpectDataFileHeaderOrLFSD
							 && expect != ExpectDataFileHeaderOrEnd) {
								cerr << "Unexpected Data File Header\n" << flush;
								return -1;
							}
							DataFileHeader *dhdr = (DataFileHeader *)Buffer;
							if (verbose) dhdr->dump(dirout);
							expect=ExpectDataFile;
						}
						else if (bytecount == 0
						      && strcmp(Buffer,"DICOMMEDIADIR") == 0) {
							if (expect != ExpectLFSD
							 && expect != ExpectDataFileHeaderOrLFSD) {
								cerr << "Unexpected Logical File System Directory\n" << flush;
								return -1;
							}
							LogicalFileSystemDirectoryFixed *lfsdf =
								(LogicalFileSystemDirectoryFixed *)Buffer;
							if (strcmp(lfsdf->UseField,"INUSE") == 0) {
								lfsdf->dump(dirout);
								unsigned long nfiles=lfsdf->TotalNumberOfFiles;
								size_t fixedsize=sizeof(LogicalFileSystemDirectoryFixed);
								const char *ptr=Buffer+fixedsize;
								size_t bytesleft=num-fixedsize;
								while (nfiles--) {
									char fnebuf[sizeof(FileNumberEntry)];
									size_t needbytes=sizeof(FileNumberEntry);
									while (needbytes > 0) {
										//cerr << "bytesleft " << bytesleft << "\n" << flush;
										//cerr << "needbytes " << needbytes << "\n" << flush;
										if (bytesleft <= 0) {
											//cerr << "bytesleft " << bytesleft << "\n" << flush;
											bytecount+=num;
											if ((num=read(64512)) > 0) {
												ptr=Buffer;
												bytesleft=num;
											}
											else {
												cerr << "Premature end of LFSD\n"
												     << flush;
												return -1;
											}
										}
										size_t addbytes=(needbytes < bytesleft)
											? needbytes
											: bytesleft;
										//cerr << "addbytes " << addbytes << "\n" << flush;
										memcpy(fnebuf,ptr,addbytes);
										ptr+=addbytes;
										bytesleft-=addbytes;
										needbytes-=addbytes;
									}
									((FileNumberEntry *)fnebuf)->dump(dirout);
								}
							}
							else {
								if (verbose) lfsdf->dump(log);
							}
							expect=ExpectDataFileHeaderOrEnd;
						}
						else {
							if (bytecount == 0 && expect != ExpectDataFile) {
								cerr << "Unexpected Data File\n" << flush;
								return -1;
							}
							expect=ExpectDataFileHeaderOrLFSD;
						}
						bytecount+=num;
						break;
				}
			}
			if (expect != ExpectDataFileHeaderOrEnd) {
				cerr << "Unexpected Data File\n" << flush;
				return -1;
			}

			if (::close(fd) == -1) return -1;

			return 0;
		}

	int Read(
			unsigned long nfiles,
			const char *const *filenames,
			const char *dicomdir,
			ostream& dirout,ostream& log,int verbose)
		{
			if (open(O_RDONLY) == -1) return -1;

			if (rewind() == -1) return -1;

			enum {
				ExpectVolumeHeader,
				ExpectLFSD,
				ExpectDataFileHeader,
				ExpectDataFile,
				ExpectDataFileHeaderOrLFSD,
				ExpectDataFileHeaderOrEnd
			} expect = ExpectVolumeHeader;

			char usefileid[72];
			usefileid[0]='\0';
			ofstream *useoutfile=0;

			int markcount=0;
			unsigned long bytecount=0;
			while (markcount < 2) {
				int num=read(64512);
				switch (num) {
					case -1:
						return -1;
					case 0:
						//if (verbose) log << "Tape Mark\n" << flush;
						if (useoutfile) {
							useoutfile->close();
							useoutfile=0;
						}
						++markcount;
						bytecount=0;
						break;
					default:
						markcount=0;
						if      (bytecount == 0
						      && strcmp(Buffer,"DICOMVOLHDR") == 0
						      && (num == sizeof(VolumeHeader)
						       || num == sizeModulo(sizeof(VolumeHeader)))
						) {
							if (expect != ExpectVolumeHeader) {
								cerr << "Unexpected Volume Header\n" << flush;
								return -1;
							}
							VolumeHeader *volhdr = (VolumeHeader *)Buffer;
							if (verbose) volhdr->dump(dirout);
							expect=ExpectLFSD;
						}
						else if (bytecount == 0
						      && strcmp(Buffer,"DICOMFILEHDR") == 0
						      && (num == sizeof(DataFileHeader)
						       || num == sizeModulo(sizeof(DataFileHeader)))
						) {
							if (expect != ExpectDataFileHeader
							 && expect != ExpectDataFileHeaderOrLFSD
							 && expect != ExpectDataFileHeaderOrEnd) {
								cerr << "Unexpected Data File Header\n" << flush;
								return -1;
							}
							DataFileHeader *dhdr = (DataFileHeader *)Buffer;
							if (verbose) dhdr->dump(dirout);
							strcpy(usefileid,dhdr->getFileID());
							expect=ExpectDataFile;
						}
						else if (bytecount == 0
						      && strcmp(Buffer,"DICOMMEDIADIR") == 0) {
							if (expect != ExpectLFSD
							 && expect != ExpectDataFileHeaderOrLFSD) {
								cerr << "Unexpected Logical File System Directory\n" << flush;
								return -1;
							}
							LogicalFileSystemDirectoryFixed *lfsdf =
								(LogicalFileSystemDirectoryFixed *)Buffer;
							if (strcmp(lfsdf->UseField,"INUSE") == 0) {
								lfsdf->dump(dirout);
								unsigned long nfiles=lfsdf->TotalNumberOfFiles;
								size_t fixedsize=sizeof(LogicalFileSystemDirectoryFixed);
								const char *ptr=Buffer+fixedsize;
								size_t bytesleft=num-fixedsize;
								while (nfiles--) {
									char fnebuf[sizeof(FileNumberEntry)];
									size_t needbytes=sizeof(FileNumberEntry);
									while (needbytes > 0) {
										//cerr << "bytesleft " << bytesleft << "\n" << flush;
										//cerr << "needbytes " << needbytes << "\n" << flush;
										if (bytesleft <= 0) {
											//cerr << "bytesleft " << bytesleft << "\n" << flush;
											bytecount+=num;
											if ((num=read(64512)) > 0) {
												ptr=Buffer;
												bytesleft=num;
											}
											else {
												cerr << "Premature end of LFSD\n"
												     << flush;
												return -1;
											}
										}
										size_t addbytes=(needbytes < bytesleft)
											? needbytes
											: bytesleft;
										//cerr << "addbytes " << addbytes << "\n" << flush;
										memcpy(fnebuf,ptr,addbytes);
										ptr+=addbytes;
										bytesleft-=addbytes;
										needbytes-=addbytes;
									}
									((FileNumberEntry *)fnebuf)->dump(dirout);
								}
							}
							else {
								if (verbose) lfsdf->dump(log);
							}
							expect=ExpectDataFileHeaderOrEnd;
						}
						else {
							if (bytecount == 0) {
								if (expect != ExpectDataFile) {
									cerr << "Unexpected Data File\n" << flush;
									return -1;
								}
								int wantit=(!nfiles || dicomdir && strcmp(usefileid,dicomdir)==0);
								int i=0;
								while (!wantit && i<nfiles) wantit|=(strcmp(usefileid,filenames[i++])==0);
								if (wantit) {
									if (verbose)
										log << usefileid << " is wanted\n" << flush;
									useoutfile= new ofstream(usefileid);
									if (!useoutfile->rdbuf()->is_open()) {
										cerr << "Can't open file <" << usefileid
										     << ">\n" << flush;
										return -1;
									}
								}
								else {
									if (verbose)
										log << usefileid << " is not wanted\n" << flush;
								}
							}
							if (useoutfile) {
								useoutfile->write(Buffer,num);
								if (!useoutfile->good()) {
									cerr << "Can't write to file <" << usefileid
									     << ">\n" << flush;
									return -1;
								}
							}
							expect=ExpectDataFileHeaderOrLFSD;
						}
						bytecount+=num;
						break;
				}
			}
			if (expect != ExpectDataFileHeaderOrEnd) {
				cerr << "Unexpected Data File\n" << flush;
				return -1;
			}

			if (::close(fd) == -1) return -1;

			return 0;
		}

	int List(
			const char *dicomdir,
			ostream& dirout,ostream& log,int verbose)
		{
			if (open(O_RDONLY) == -1) return -1;

			if (rewind() == -1) return -1;

			enum {
				ExpectVolumeHeader,
				ExpectLFSD,
				ExpectDataFileHeader,
				ExpectDataFile,
				ExpectDataFileHeaderOrLFSD,
				ExpectDataFileHeaderOrEnd
			} expect = ExpectVolumeHeader;
			int markcount=0;
			unsigned long bytecount=0;
			while (markcount < 2) {
				int num=read(64512);
				switch (num) {
					case -1:
						return -1;
					case 0:
						//if (verbose) log << "Tape Mark\n" << flush;
						++markcount;
						bytecount=0;
						break;
					default:
						markcount=0;
						if      (bytecount == 0
						      && strcmp(Buffer,"DICOMVOLHDR") == 0
						      && (num == sizeof(VolumeHeader)
						       || num == sizeModulo(sizeof(VolumeHeader)))
						) {
							if (expect != ExpectVolumeHeader) {
								cerr << "Unexpected Volume Header\n" << flush;
								return -1;
							}
							VolumeHeader *volhdr = (VolumeHeader *)Buffer;
							if (verbose) volhdr->dump(dirout);
							expect=ExpectLFSD;
						}
						else if (bytecount == 0
						      && strcmp(Buffer,"DICOMFILEHDR") == 0
						      && (num == sizeof(DataFileHeader)
						       || num == sizeModulo(sizeof(DataFileHeader)))
						) {
							if (expect != ExpectDataFileHeader
							 && expect != ExpectDataFileHeaderOrLFSD
							 && expect != ExpectDataFileHeaderOrEnd) {
								cerr << "Unexpected Data File Header\n" << flush;
								return -1;
							}
							DataFileHeader *dhdr = (DataFileHeader *)Buffer;
							if (verbose) dhdr->dump(dirout);
							expect=ExpectDataFile;
						}
						else if (bytecount == 0
						      && strcmp(Buffer,"DICOMMEDIADIR") == 0) {
							if (expect != ExpectLFSD
							 && expect != ExpectDataFileHeaderOrLFSD) {
								cerr << "Unexpected Logical File System Directory\n" << flush;
								return -1;
							}
							LogicalFileSystemDirectoryFixed *lfsdf =
								(LogicalFileSystemDirectoryFixed *)Buffer;
							if (strcmp(lfsdf->UseField,"INUSE") == 0) {
								lfsdf->dump(dirout);
								unsigned long nfiles=lfsdf->TotalNumberOfFiles;
								size_t fixedsize=sizeof(LogicalFileSystemDirectoryFixed);
								const char *ptr=Buffer+fixedsize;
								size_t bytesleft=num-fixedsize;
								while (nfiles--) {
									char fnebuf[sizeof(FileNumberEntry)];
									size_t needbytes=sizeof(FileNumberEntry);
									while (needbytes > 0) {
										//cerr << "bytesleft " << bytesleft << "\n" << flush;
										//cerr << "needbytes " << needbytes << "\n" << flush;
										if (bytesleft <= 0) {
											//cerr << "bytesleft " << bytesleft << "\n" << flush;
											bytecount+=num;
											if ((num=read(64512)) > 0) {
												ptr=Buffer;
												bytesleft=num;
											}
											else {
												cerr << "Premature end of LFSD\n"
												     << flush;
												return -1;
											}
										}
										size_t addbytes=(needbytes < bytesleft)
											? needbytes
											: bytesleft;
										//cerr << "addbytes " << addbytes << "\n" << flush;
										memcpy(fnebuf,ptr,addbytes);
										ptr+=addbytes;
										bytesleft-=addbytes;
										needbytes-=addbytes;
									}
									((FileNumberEntry *)fnebuf)->dump(dirout);
								}
							}
							else {
								if (verbose) lfsdf->dump(log);
							}
							expect=ExpectDataFileHeaderOrEnd;
						}
						else {
							if (bytecount == 0 && expect != ExpectDataFile) {
								cerr << "Unexpected Data File\n" << flush;
								return -1;
							}
							expect=ExpectDataFileHeaderOrLFSD;
						}
						bytecount+=num;
						break;
				}
			}
			if (expect != ExpectDataFileHeaderOrEnd) {
				cerr << "Unexpected Data File\n" << flush;
				return -1;
			}

			if (::close(fd) == -1) return -1;

			return 0;
		}
#endif // USEDIRECTORYACCESS
};

static char *
ValidateDicomFileIDs(int nfiles,const char *const *filenames,int reserved)
{
	ostrstream ostr;
	if (nfiles) {
		const char *const *f = filenames;
		while (nfiles--) {
			const char *p = *f;
			int delimitercount=0;
			int componentlength=0;
			int bytecount=0;
			char c;
			while ((c=*p++)) {
				++bytecount;
				if (c == '\\') {
					if (bytecount <= 1) {
						ostr << "Bad Dicom FileID: "
						     << "leading backslash illegal "
						     << "in <" << *f << ">\n";
						break;
					}
					if (componentlength < 1) {
						ostr << "Bad Dicom FileID: "
						     << "empty components illegal "
						     << "in <" << *f << ">\n";
						break;
					}
					componentlength=0;
					if (++delimitercount > 7) {
						ostr << "Bad Dicom FileID: "
						     << "too many components (> 8) "
						     << "in <" << *f << ">\n";
						break;
					}
				}
				else if (isupper(c) || isdigit(c) || c == '_') {
					if (++componentlength > 8) {
						ostr << "Bad Dicom FileID: "
						     << "component too long (> 8 characters) "
						     << "in <" << *f << ">\n";
						break;
					}
				}
				else {
					if (islower(c)) {
						ostr << "Bad Dicom FileID: "
						     << "only uppercase allowed "
						     << "in <" << *f << ">\n";
					}
					else {
						ostr << "Bad Dicom FileID: "
						     << "illegal character <" << c << "> "
						     << "in <" << *f << ">\n";
					}
					break;
				}
			}
			//if (bytecount > (8*8+7)) {
			//	ostr << "Bad Dicom FileID: "
			//	     << "total ID length too long at "
			//	     << bytecount << " characters "
			//	     << "in <" << *f << ">\n";
			//}
			if (reserved && strcmp(*f,"DICOMDIR") == 0) {
				ostr << "Bad Dicom FileID: "
				     << "illegal use of reserved name "
				     << "in <" << *f << ">\n";
			}
			++f;
		}
	}
	if (ostr.pcount()) {
		return ostr.str();
	}
	else {
		return 0;
	}
}

int
main(int argc,char **argv)
{
	//CCOMMAND(argc,argv);

	//cerr.setf(ios::showbase,ios::showbase);

	char *device	= "/dev/nrst0";
	char *kind	= 0;
	char *dicomdir	= 0;
	int verbose	= 0;
	int dontusedirectory = 0;
	char mode	= 0;

	GetOpt getopt(argc,argv,"cd:f:k:lruvn");
	int c;
	int err=0;
	while ((c=getopt()) != -1 && !err) {
		switch (c) {
			case 'f':	// tape device
				device=getopt.optarg;
				break;
			case 'k':	// kind of device
				kind=getopt.optarg;
				break;
			case 'd':	// DICOMDIR name
				dicomdir=getopt.optarg;
				break;
			case 'v':	// verbose
				verbose=1;
				break;
			case 'n':	// no directory
				dontusedirectory=1;
				break;
			case 'c':	// Create
			case 'u':	// Update
			case 'r':	// Read
			case 'l':	// List/extract directory
				mode=c;
				break;
			case '?':
			default:
				cerr << "Illegal option <" << (char)c 
				     << ">\n" << flush;
				err=1;
		}
	}

	if (!mode) {
		cerr << "Must specify -c(create),-u(update),-r(read), or "
		     << "-l(list/extract directory)\n" << flush;
		err=1;
	}

	size_t fixedblocklength;
	size_t lastblockmodulo;

	if (kind) {
		if (strcmp(kind,"dat") == 0) {
			fixedblocklength=64512;
			lastblockmodulo=0;
		}
		else if (strcmp(kind,"8mm") == 0) {
			fixedblocklength=64512;
			lastblockmodulo=1024;
		}
		else if (strcmp(kind,"qic") == 0) {
			fixedblocklength=64512;
			lastblockmodulo=512;
		}
		else if (strcmp(kind,"9track") == 0) {
			fixedblocklength=8192;
			lastblockmodulo=0;
		} else {
			cerr << "Bad -k kind of device\n" << flush;
			err=1;
		}
	}
	else {
		cerr << "Must specify -k kind of device\n" << flush;
		err=1;
	}

	int nfiles=argc-getopt.optind;
	const char *const *filenames;

	if (nfiles == 0 && (mode == 'c' || mode == 'u')) {
		cerr << "Must have filename list if creating or updating\n"
		     << flush;
		err=1;
	}
	else if (nfiles > 0 &&  mode == 'l') {
		cerr << "Must not have filename list if listing\n"
		     << flush;
		err=1;
	}
	else {
		filenames=&argv[getopt.optind];
	}

	if (nfiles) {
		char *s=ValidateDicomFileIDs(nfiles,filenames,1);
		if (s) {
			cerr << s << flush;
			delete[] s;
			err=1;
		}
	}

	if (err) {
		cerr << "Usage: " << argv[0]
		     << " -c|u|r|l -f device -k dat|8mm|qic|9track [-v] [-n]"
		     << " [-d DICOMDIR] [filenames]\n" << flush;
		return 1;
	}

	if (verbose) {
		switch (mode) {
			case 'c':
				cerr << "Creating on ";
				break;
			case 'u':
				cerr << "Updating on ";
				break;
			case 'r':
				cerr << "Reading from ";
				break;
			case 'l':
				cerr << "Listing from ";
				break;
		}
		cerr << "device " << device
		     << " (" << kind << ")"
		     << ", blocksize " << fixedblocklength << " bytes"
		     << ", last block padded to " << lastblockmodulo << " bytes"
		     << ", " << (dontusedirectory ? "don't " : " ") << "use directory"
		     << "\n" << flush;
	}

	SequentialMediaFileSystem d(device,fixedblocklength,lastblockmodulo);

	switch (mode) {
		case 'c':
			d.Create(nfiles,filenames,dicomdir,cerr,verbose,dontusedirectory);
			break;
		case 'u':
			d.Update(nfiles,filenames,dicomdir,cerr,verbose,dontusedirectory);
			break;
		case 'r':
			d.ReadOrList(nfiles,filenames,dicomdir,cout,cerr,verbose,1,0,dontusedirectory);
			break;
		case 'l':
			d.ReadOrList(0,0,dicomdir,cout,cerr,verbose,0,1,dontusedirectory);
			break;
	}

	return 0;
}
