static char *CopyrightIdentifier(void) { return "@(#)ge98todc.cc 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved."; }

#ifdef __GNUG__
#include <builtin.h>
#endif
#include <math.h>
#include <string.h>

#include <strstream.h>
#include <iomanip.h>

#include "tlist.h"
#include "tfstream.h"
#include "tddictc.h"

#include "dgtypes.h"

#include "GetOpt.h"
#include "MainCmd.h"

static ifstream	*instream;

static Uint16 	width;
static Uint16 	height;
static Uint16 	usemap;
static Uint16 	mapblock;
static Uint16 	datablock;
static Uint16	*map;
static Uint16 	currentrow;
static Uint16 	*linebuffer;
static Int16	last_pixel;

void
read_header(
	ifstream& instream,		// where to read from
	DC3ListofAttributes& list,	// list to add new attributes to
	const char *stamp,		// substitute date/time stamp for UIDs
	Uint16& width,			// Returned - image width
	Uint16& height,			// Returned - image height
	Uint16& usemap,			// Returned - use map
	Uint16& mapblock,		// Returned - block offset to map
	Uint16& datablock		// Returned - block offset to image
);

static Uint16
readUint16(ifstream *instream)
{
	Uint16 u;
	unsigned char buffer[2];
	instream->read(buffer,2);
	// DataGeneral is a Big Endian machine
	u =  (Uint16)buffer[0];
	u <<= 8;
	u |= (Uint16)buffer[1];
	return u;
}

static streampos
seekblock(ifstream *instream,unsigned block)
{
	const unsigned offset = block*256*2;	// 256 word blocks,16 bit words
	return instream->rdbuf()->seekoff
			(offset,ios::beg,ios::in|ios::out);
}

static int
read9800linestart(void)
{
	// instream is already open

	instream->clear();	// Clear EOF state to allow seeking/reading

	// Find the start of the perimeter encoding map ...

	if (usemap) {
		if (seekblock(instream,mapblock) == streampos(EOF)) {
			cerr << "Map seek failed\n" << flush;
			return -1;
		}

		// Allocate & read the map ...

		map=new Uint16[height];
		if (!map) {
			cerr << "Couldn't allocate map\n" << flush;
			return -1;
		}

		unsigned i;
		for (i=0; i<height; ++i) {
			map[i]=readUint16(instream);
		}

		if (!instream->good()) {
			cerr << "Map read failed\n" << flush;
			return -1;
		}
	}
	else {
		map=0;
	}

	// Position at start of image pixel data ...

	if (seekblock(instream,datablock) == streampos(EOF)) {
		cerr << "Image seek failed\n" << flush;
		return -1;
	}

	// Initialize decompression parameters ...

	currentrow=0;		// used as index into map
	last_pixel=0;		// used for DPCM decoding

	linebuffer=new Uint16[width];
	if (!linebuffer) {
		cerr << "Couldn't allocate linebuffer\n" << flush;
		return -1;
	}

	return 0;
}

static int
read9800linedone(void)
{
	// leave instream open - closed in main routine

	if (linebuffer)	delete [] linebuffer;
	if (map)	delete [] map;

	return 0;
}

static int
read9800line(Uint16 **buffer)
{
	Uint16 		*ptr 		= linebuffer;
	unsigned 	start;
	unsigned 	end;
	unsigned 	j;

	if (map) {
		unsigned line	= map[currentrow];
		start	= width/2-line;
		end	= start+line*2;
	}
	else {
		start	= 0;
		end	= width;
	}

	// Pad the first "empty" part of the line ...
	for (j=0; j<start; j++) *ptr++=0;

	// Copy the middle of the line (compressed or uncompressed)
	while (start<end) {
		unsigned char byte;
		instream->read(&byte,1);
		if (!instream->good()) break;
		if (byte & 0x80) {
			signed char delta;
			if (byte & 0x40) {
				delta=byte;
			}
			else {
		    		delta=byte & 0x3f;
			}
			last_pixel+=delta;
		}
		else {
			last_pixel=byte << 8;
			instream->read(&byte,1);
			if (!instream->good()) break;
			last_pixel+=byte;
		}
		*ptr++=(Uint16)last_pixel & 0x0fff;
		++start;
	}

	// Pad the last "empty" part of the line ...
	for (j=end; j<width; j++) *ptr++=0;

	++currentrow;

	if (!instream->good()) {
		cerr << "Line read failed\n" << flush;
		return -1;
	}

	*buffer=linebuffer;		// return pointer to data read
	return width;			// return length of data (1 line)
}

int
main(int argc,char *argv[])
{
	CCOMMAND(argc,argv);

	char *	stamp		= 0;
	char *	inname		= 0;
	char *	outname		= 0;
	int 	verbose		= 0;
	int 	metaheader 	= 0;
	char *	outsyntaxUID	= DC3DefaultTransferSyntax;

	GetOpt getopt(argc,argv,"d:mo:r[2]s:v");
	int c;
	int err=0;
	DC3ListofAttributes deletelist;
	DC3ListofAttributes replacelist;
	while ((c=getopt()) != -1 && !err) {
		switch (c) {
			case 'v':	// Verbose
				verbose=1;
				break;
			case 'd':	// Delete attribute keyword
				{
					char *keyword=getopt.optarg;
					char *values="";
					DC3Attribute *attr;
					DC3AttributeGet(keyword,values,
						&attr,&err);
					if (!attr || err) {
						cerr << "Couldn't handle <"
						     << keyword
						     << "> attribute\n"
						     << flush;
					}
					else {
						deletelist.insert(attr);
					}
				}
				break;
			case 'r':	// Replace attribute keyword & value
				{
					char *keyword=getopt.optarg;
					char *values=argv[getopt.optind];
					DC3Attribute *attr;
					DC3AttributeGet(keyword,values,
						&attr,&err);
					if (!attr || err) {
						cerr << "Couldn't handle <"
						     << keyword
						     << "> attribute\n"
						     << flush;
					}
					else {
						replacelist.insert(attr);
					}
				}
				break;
			case 'm':	// Generate metaheader
				metaheader=1;
				break;
			case 'o':	// Output transfer syntax
				outsyntaxUID=getopt.optarg;
				break;
			case 's':	// Stamp with specified date/time
				stamp=getopt.optarg;
				break;
			case '?':
			default:
				err=1;
		}
	}

	if (getopt.optind+2 == argc) {
		inname=argv[getopt.optind];
		outname=argv[getopt.optind+1];
	}
	else err=1;

	if (err) {
		cerr << "Usage: " << argv[0] << " [-m] [-o syntax] "
		     << "[-d key] [-r key value] [-s stamp] [-v] "
		     << "infile outfile\n"
		     << flush;
		return 1;
	}

	if (verbose) {
		cerr << "************** Deletions ************\n" << flush;
		deletelist.first();
		while (deletelist.ismore()) {
			DC3Attribute *attr = deletelist.value();
			attr->put(cerr);
			deletelist.next();
		}
		cerr << "************** Replacements ************\n" << flush;
		replacelist.first();
		while (replacelist.ismore()) {
			DC3Attribute *attr = replacelist.value();
			attr->put(cerr);
			replacelist.next();
		}
	}

	instream = new ifstream(inname,ios::in);
	if (!instream->rdbuf()->is_open()) {
		cerr << "Can't open <" << inname << "> for read\n" << flush;
		return 1;
	}

	DC3ListofAttributes list;

	if (verbose)
		cerr << "**************** Reading **************\n" << flush;

	read_header(*instream,list,stamp,
		width,height,usemap,mapblock,datablock);

	{
		// Add a PixelData attribute ...

		DC3Tag tag (DC3PixelData_GROUP,DC3PixelData_ELEMENT);
		DC3OtherWordString *attr = 
			new DC3OtherWordString(tag,
				width*height*2,
				read9800linestart,
				read9800line,
				read9800linedone);
		list.insert(attr);
	}

	if (verbose)
		cerr << "**************** Deleting **************\n" << flush;

	list.Delete(deletelist);

	if (verbose)
		cerr << "**************** Replacing **************\n" << flush;

	list.Replace(replacelist);

	if (verbose)
		cerr << "**************** Fixing **************\n" << flush;

	list.Fix(outsyntaxUID,metaheader);

	if (verbose)
		cerr << "**************** Writing **************\n" << flush;

	DC3TransferSyntax outtest(outsyntaxUID);
	if (!outtest.isValid()) {
		cerr << "Output Transfer Syntax UID <"
		     << outsyntaxUID << "> is invalid\n";
		return 1;
	}

	DC3ofstream outstream(outname,outsyntaxUID);
	if (!outstream.is_open()) {
		cerr << "Can't open <" << outname << "> for write\n" << flush;
		return 1;
	}

	list.first();
	while (list.ismore()) {
		DC3Attribute *attr = list.value();
		if (verbose) {
			attr->put(cerr);
		}
		attr->write(outstream);
		list.next();
	}

	outstream.close();

	return 0;
}

