%{

#ifndef YYBISON
#include <stdio.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

struct list {
	int count;
	char *string;
	struct list *next;
};

int g_lineno = 1;

#include "s15exnmr.h"

ExtractedNMRParameters enmr;

void		listtoarray	(struct list *list, char **array, int n);
struct list *	mergelist	(struct list *list1,struct list *list2);
struct list *	newlist		(int n,char *string);

%}

%union {
	char *string;
	int none;
	struct list *list;
}

%token	<string>	STRING INTEGER REAL

%token	<none>		COLON COMMA OPENPAREN CLOSEPAREN

%token	<none>		NUCLEUS
%token	<none>		SCANMODE
%token	<none>		SEQUENCE
%token	<none>		THICKNESS
%token	<none>		OFFCENTER
%token	<none>		FOV
%token	<none>		RESOLUTION
%token	<none>		SLICES3D
%token	<none>		SLICESMS
%token	<none>		SLICEFACTOR
%token	<none>		MEASUREMENTS
%token	<none>		ECHOS
%token	<none>		TE1
%token	<none>		TE2
%token	<none>		TRSE
%token	<none>		TRIR
%token	<none>		TI
%token	<none>		ORIENTM
%token	<none>		ORIENTP
%token	<none>		ORIENTS
%token	<none>		PRESATSPECTRAL
%token	<none>		PRESAT
%token	<none>		FLOWCOMP
%token	<none>		COILTX
%token	<none>		COILRX
%token	<none>		WEIGHT
%token	<none>		BIRTHYEAR
%token	<none>		FREQSYNTH
%token	<none>		FLIPANGLE
%token	<none>		BWPERPIXEL

%token	<none>		ID

%type	<list>		valuelist
%type	<list>		integerentry integerlist
%type	<list>		realentry reallist
%type	<list>		stringentry stringlist

%start commandlist

%%

commandlist:	commandline
	|	commandlist commandline
	;

commandline:	ID COLON valuelist			{}
	|	NUCLEUS COLON STRING			{ enmr.nucleus=$3; }
	|	SCANMODE COLON STRING			{ enmr.scanmode=$3; }
	|	SEQUENCE COLON STRING			{ enmr.sequence=$3; }
	|	THICKNESS COLON reallist		{ listtoarray($3,&enmr.thickness,1); }
	|	OFFCENTER COLON reallist		{ listtoarray($3,enmr.offcenter,3); }
	|	FOV COLON reallist			{ listtoarray($3,enmr.fov,2); }
	|	RESOLUTION COLON integerlist		{ listtoarray($3,enmr.resolution,2); }
	|	SLICES3D COLON integerlist		{ listtoarray($3,&enmr.slices3d,1); }
	|	SLICESMS COLON integerlist		{ listtoarray($3,&enmr.slicesms,1); }
	|	SLICEFACTOR COLON reallist		{ listtoarray($3,&enmr.slicefactor,1); }
	|	ECHOS COLON integerlist			{ listtoarray($3,&enmr.echos,1); }
	|	MEASUREMENTS COLON integerlist		{ listtoarray($3,&enmr.measurements,1); }
	|	TE1 COLON reallist			{ listtoarray($3,&enmr.te1,1); }
	|	TE2 COLON reallist			{ listtoarray($3,&enmr.te2,1); }
	|	TRSE COLON reallist			{ listtoarray($3,&enmr.trse,1); }
	|	TRIR COLON reallist			{ listtoarray($3,&enmr.trir,1); }
	|	TI COLON reallist			{ listtoarray($3,&enmr.ti,1); }
	|	ORIENTM COLON reallist			{ listtoarray($3,enmr.orientm,3); }
	|	ORIENTP COLON reallist			{ listtoarray($3,enmr.orientp,3); }
	|	ORIENTS COLON reallist			{ listtoarray($3,enmr.orients,3); }
	|	PRESATSPECTRAL COLON STRING		{ enmr.presatspectral=$3; }
	|	PRESAT COLON STRING			{ enmr.presat=$3; }
	|	FLOWCOMP COLON STRING			{ enmr.flowcomp=$3; }
	|	COILTX COLON STRING			{ enmr.coiltx=$3; }
	|	COILRX COLON STRING			{ enmr.coilrx=$3; }
	|	WEIGHT COLON reallist			{ listtoarray($3,&enmr.weight,1); }
	|	BIRTHYEAR COLON integerlist		{ listtoarray($3,&enmr.birthyear,1); }
	|	FREQSYNTH COLON integerlist		{ listtoarray($3,&enmr.freqsynth,1); }
	|	FLIPANGLE COLON integerlist		{ listtoarray($3,&enmr.flipangle,1); }
	|	BWPERPIXEL COLON reallist		{ listtoarray($3,&enmr.bwperpixel,1); }
	;

valuelist:	integerlist
	|	reallist
	|	stringlist
	;

integerlist:	integerentry				{ $$=$1; }
	|	integerlist COMMA integerentry		{ $$=mergelist($1,$3); }
	;

integerentry:	INTEGER					{ $$=newlist(1,$1); }
	|	OPENPAREN INTEGER CLOSEPAREN INTEGER	{ $$=newlist(atoi($2),$4); }
	;

reallist:	realentry				{ $$=$1; }
	|	reallist COMMA realentry		{ $$=mergelist($1,$3); }
	;

realentry:	REAL					{ $$=newlist(1,$1); }
	|	OPENPAREN INTEGER CLOSEPAREN REAL	{ $$=newlist(atoi($2),$4); }
	;

stringlist:	stringentry				{ $$=$1; }
	|	stringlist COMMA stringentry		{ $$=mergelist($1,$3); }
	;

stringentry:	STRING					{ $$=newlist(1,$1); }
	;

%%

int
yyerror(char *s)	/* in case not supplied (eg. bison) */
{ 
	fprintf (stderr, "%s\n", s); 
}

void 
listtoarray(struct list *list, char **array, int n)
{
	struct list *p = list;
	int oldn=n;
	char **oldarray=array;
	while (p) {
		int count=p->count;
		while (count-- > 0 && n-- > 0) {
			*array++=p->string;
		}
		p=p->next;
	}
}

struct list *
mergelist(struct list *list1,struct list *list2)
{
	/* add list2 on end of list1 */
	struct list *p = list1;
	while (p && p->next) p=p->next;
	if (p) p->next=list2;
	return list1;
}

struct list *
newlist(int n,char *string)
{
	struct list *p = (struct list *) malloc(sizeof(struct list));
	if (p) {
		p->count=n;
		p->string=string;
	}
	return p;
}

extern int yydebug;

ExtractedNMRParameters *
getExtractedNMRParameters(void)
{
	zeroExtractedNMRParameters(enmr);

	/* yydebug=1; */
	yyparse();

	return &enmr;
}

#ifdef S15EXNMRMAIN
int
main(int argc,char **argv)
{
	char *tr=0;
	char *ti=0;
	char *slices=0;

	ExtractedNMRParameters *e = getExtractedNMRParameters();

	/* dump the parsed values to stderr ... */

	/* why bother ? ... */
	/* unfolding the parentheses makes for easier use in scripts */

	if (e->nucleus) fprintf(stderr,"nucleus : \"%s\"\n",e->nucleus);
	if (e->scanmode) fprintf(stderr,"scan_mode : \"%s\"\n",e->scanmode);
	if (e->sequence) fprintf(stderr,"imagine->sequence : \"%s\"\n",e->sequence);

	if (e->thickness) fprintf(stderr,"slice_thickness : %s\n",e->thickness);

	if (e->offcenter[0]) fprintf(stderr,"off_center_distance : %s, %s, %s\n",
		e->offcenter[0],e->offcenter[1],e->offcenter[2]);
	if (e->fov[0]) fprintf(stderr,"field_of_view : %s, %s\n",
		e->fov[0],e->fov[1]);
	if (e->resolution[0]) fprintf(stderr,"scan_resolution : %s, %s\n",
		e->resolution[0],e->resolution[1]);

	if (e->slices3d) fprintf(stderr,"slices_3D : %s\n",e->slices3d);
	if (e->slicesms) fprintf(stderr,"slices_MS : %s\n",e->slicesms);
	if (e->slicefactor) fprintf(stderr,"slice_factor : %s\n",e->slicefactor);
	if (e->measurements) fprintf(stderr,"measurements : %s\n",e->measurements);
	if (e->echos) fprintf(stderr,"echos : %s\n",e->echos);
	if (e->te1) fprintf(stderr,"first_echo_time : %s\n",e->te1);
	if (e->te2) fprintf(stderr,"second_echo_time : %s\n",e->te2);
	if (e->trse) fprintf(stderr,"SE_repetition_time : %s\n",e->trse);
	if (e->trir) fprintf(stderr,"IR_repetition_time : %s\n",e->trir);
	if (e->ti) fprintf(stderr,"IR_delay : %s\n",e->ti);

	if (e->orientm[0]) fprintf(stderr,"m_orientation : %s, %s, %s\n",
		e->orientm[0],e->orientm[1],e->orientm[2]);
	if (e->orientp[0]) fprintf(stderr,"p_orientation : %s, %s, %s\n",
		e->orientp[0],e->orientp[1],e->orientp[2]);
	if (e->orients[0]) fprintf(stderr,"s_orientation : %s, %s, %s\n",
		e->orients[0],e->orients[1],e->orients[2]);

	if (e->presatspectral) fprintf(stderr,"spectral_presaturation : \"%s\"\n",e->presatspectral);
	if (e->presat) fprintf(stderr,"enable_presat : \"%s\"\n",e->presat);
	if (e->flowcomp) fprintf(stderr,"flow_comp : \"%s\"\n",e->flowcomp);
	if (e->coiltx) fprintf(stderr,"send_coil : \"%s\"\n",e->coiltx);
	if (e->coilrx) fprintf(stderr,"receive_coil : \"%s\"\n",e->coilrx);

	if (e->weight) fprintf(stderr,"patient_weight : %s\n",e->weight);
	if (e->birthyear) fprintf(stderr,"patient_birth_year : %s\n",e->birthyear);
	if (e->freqsynth) fprintf(stderr,"synthesizer_frequency : %s\n",e->freqsynth);
	if (e->flipangle) fprintf(stderr,"angleRF_partial_flip : %s\n",e->flipangle);
	if (e->bwperpixel) fprintf(stderr,"actual_BW_per_pixel : %s\n",e->bwperpixel);

	/* write the DICOM attribute commands to stdout ... */

	fprintf(stdout," -r ImageType \"ORIGINAL\\PRIMARY\\MPR\" ");
	fprintf(stdout," -r SOPClassUID \"1.2.840.10008.5.1.4.1.1.4\" ");
	fprintf(stdout," -r Modality MR ");
	fprintf(stdout," -r Manufacturer PHILIPS ");
	fprintf(stdout," -r ManufacturerModelName GYROSCAN ");

	/* if (e->birthyear)
		fprintf(stdout," -r PatientBirthDate \"%s0000\" ",e->birthyear); */
	if (e->weight)
		fprintf(stdout," -r PatientWeight \"%E\" ",atof(e->weight)*1000.0);

	if (e->sequence) {
		if (strcmp(e->sequence,"SE") == 0) {
			fprintf(stdout," -r ScanningSequence SE ");
			tr=e->trse;
		}
		else if (strcmp(e->sequence,"IR") == 0) {
			fprintf(stdout," -r ScanningSequence IR ");
			tr=e->trir;
			ti=e->ti;
		}
	}
	fprintf(stdout," -r SequenceVariant NONE ");

	{
		if (e->presatspectral && strcmp(e->presatspectral,"no") == 0)
			e->presatspectral=0;
		if (e->presat && strcmp(e->presat,"no") == 0)
			e->presat=0;
		if (e->flowcomp && strcmp(e->flowcomp,"no") == 0)
			e->flowcomp=0;
		fprintf(stdout," -r ScanOptions \"");
		if (e->presatspectral) {
			fprintf(stdout,"FS");
			if (e->presat || e->flowcomp) fprintf(stdout,"\\");
		}
		if (e->presat) {
			fprintf(stdout,"SP");
			if (e->flowcomp) fprintf(stdout,"\\");
		}
		if (e->flowcomp) fprintf(stdout,"FC");
		fprintf(stdout,"\" ");
	}

	if (e->scanmode && strcmp(e->scanmode,"MS") == 0) {
		fprintf(stdout," -r MRAcquisitionType 2D ");
		slices=e->slicesms;
	}

	if (e->thickness)
		fprintf(stdout," -r SliceThickness \"%s\" ",e->thickness);

	if (tr)
		fprintf(stdout," -r RepetitionTime \"%s\" ",tr);
	if (ti)
		fprintf(stdout," -r InversionTime \"%s\" ",ti);

	if (e->measurements)
		fprintf(stdout," -r NumberOfAverages \"%s\" ",e->measurements);

	if (e->freqsynth)
		fprintf(stdout," -r ImagingFrequency \"%s\" ",e->freqsynth);
	if (e->nucleus)
		fprintf(stdout," -r ImagedNucleus \"%s\" ",e->nucleus);

	if (e->thickness && e->slicefactor)
		fprintf(stdout," -r SpacingBetweenSlices \"%E\" ",
			atof(e->thickness)*(atof(e->slicefactor)-1.0));

	if (e->bwperpixel)
		fprintf(stdout," -r PixelBandwidth \"%s\" ",e->bwperpixel);

	if (e->fov[0] && e->fov[1])
		fprintf(stdout," -r ReconstructionDiameter \"%E\" ",
			(atof(e->fov[0])+atof(e->fov[1]))/2.0);

	if (e->coilrx)
		fprintf(stdout," -r ReceivingCoil \"%s\" ",e->coilrx);
	if (e->coiltx)
		fprintf(stdout," -r TransmittingCoil \"%s\" ",e->coiltx);

	if (e->resolution[0] && e->resolution[1])
		fprintf(stdout," -r AcquisitionMatrix \"%s\\%s\" ",
			e->resolution[0],e->resolution[1]);

	if (e->flipangle)
		fprintf(stdout," -r FlipAngle \"%s\" ",e->flipangle);

/* char *e->offcenter[3] = { 0, 0, 0 };		*/
/* char *e->slices3d = 0;			*/
/* char *e->te1 = 0;ms EchoTime			*/
/* char *e->te2 = 0;ms EchoTime			*/
/* char *e->orientm[3] = { 0, 0, 0 };		*/
/* char *e->orientp[3] = { 0, 0, 0 };		*/
/* char *e->orients[3] = { 0, 0, 0 };		*/

	exit(0);
}

#endif /* S15EXNMRMAIN */
