static char *CopyrightIdentifier(void) { return "@(#)s15todc.cc 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved."; }
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "tlist.h"
#include "tfstream.h"
#include "tddictc.h"
#include "tuidgen.h"
#include "tabstc.h"
#include "tattrmk.h"

#include "GetOpt.h"
#include "MainCmd.h"

#include "s15exnmr.h"
#include "s15ptdrc.h"

static unsigned width;
static unsigned height;
static ifstream *instream;
static Uint16 	*linebuffer;

static Uint16
readUint16(ifstream *instream)
{
	Uint16 u;
	unsigned char buffer[2];
	instream->read(buffer,2);
	// S15 REC file is Little Endian
	u =  (Uint16)buffer[1];
	u <<= 8;
	u |= (Uint16)buffer[0];
	return u;
}

static int
reads15linestart(void)
{
	// instream is already open

	// Initialize parameters ...

	linebuffer=new Uint16[width];
	if (!linebuffer) {
		cerr << "Couldn't allocate linebuffer\n" << flush;
		return -1;
	}

	return 0;
}

static int
reads15linedone(void)
{
	// leave instream open - closed in main routine

	if (linebuffer)	delete [] linebuffer;

	return 0;
}

static int
reads15line(Uint16 **buffer)
{
	Uint16 *ptr = linebuffer;

	unsigned col;
	for (col=0; col<width; ++col) {
		Uint16 u=readUint16(instream);
		if (!instream->good()) break;
		*ptr++=u;
	}

	if (!instream->good()) {
		cerr << "Line read failed\n" << flush;
		return -1;
	}

	*buffer=linebuffer;		// return pointer to data read
	return width;			// return length of data (1 line)
}

// should return const but this messes with current
// revision of attr constructors ... leave for now :(

static /*const*/ char *
extract(const char *src,unsigned length)
{
	static char buffer[80];
	strncpy(buffer,src,length);
	buffer[length]=0;
	return buffer;
}

static void
getDrcDate(const char *src,int length,unsigned &dd,unsigned &mm,unsigned &yyyy)
{
	cerr << "getDrcDate: extract <" << extract(src,length) << ">\n";
	dd=0; mm=0; yyyy=0;
	while (length>0 && isspace(*src)) { --length; ++src; }
	while (length>0 && isdigit(*src)) { --length; dd=dd*10+(*src++-'0'); }
	while (length>0 && isspace(*src)) { --length; ++src; }
	if (--length>2) {
		++src;	// skip '-'
		while (length>0 && isspace(*src)) { --length; ++src; }
		switch (tolower(*src)) {
			case 'j': 	mm=(*(src+1) == 'a')
						? 1		// jan
						: ((*(src+2) == 'n')
							? 6 	// jun
							: 7 	// jul
							)
						;
					break;
			case 'f':	mm=2;  break;	// feb
			case 'm': 	mm=(*(src+2) == 'r')
						? 3	// mar
						: 5 	// may
						;
					break;
			case 'a':	mm=(*(src+1) == 'p')
						? 4	// apr
						: 8 	// aug
						;
					break;
			case 's':	mm=9;  break;	// sep
			case 'o':	mm=10; break;	// oct
			case 'n':	mm=11; break;	// nov
			case 'd':	mm=12; break;	// dec
			default:	break;
		}
		length+=3;
	}
	while (length>0 && isspace(*src)) { --length; ++src; }
	if (--length>0) {
		++src;	// skip '-'
		while (length>0 && isspace(*src)) { --length; ++src; }
		while (length>0 && isdigit(*src)) { --length; yyyy=yyyy*10+(*src++-'0'); }
		if (yyyy < 100) yyyy+=(yyyy < 70 ? 2000 : 1900);
	}	
}

static int
getExtractedPatientDrcParameters(const char *drcname,InitialString *&is,FixedFields *&ff)
{
	ifstream drcin(drcname);

	if (!drcin.rdbuf()->is_open()) {
		cerr << "Can't open " << drcname << " for read\n";
		return 1;
	}

	const size_t buflng = 512;
	char *buffer = new char[buflng];
	drcin.read(buffer,buflng);
	unsigned count=drcin.gcount();
	if (!drcin.good() || count < 4) {
		cerr << "Can't read from " << drcname << "\n";
		return 1;
	}

	is=(InitialString *)buffer;
	unsigned islength=atoi(is->length);
	ff=(FixedFields *)(buffer+islength);

	if (islength+FixedFieldsLength > buflng
	 || islength+FixedFieldsLength > count
	 || islength != InitialStringLength
	 || atoi(ff->length) != FixedFieldsLength
	) {
		cerr << "Internal Error - "
		     << "getExtractedPatientDrcParameters: length wrong: "
		     << " islength=" << islength
		     << " (wanted " << InitialStringLength
		     << ") fflength=" << atoi(ff->length)
		     << " (wanted " << FixedFieldsLength
		     << "), buflng=" << buflng
		     << ", count=" << count
		     << "\n";
		cerr << "Internal Error - please inform author\n";
		return 1;
	}


	return 0;
}

int
main(int argc,char **argv)
{
	CCOMMAND(argc,argv);

	char *	stamp		= 0;
	int	study		= 0;
	int	series		= 0;
	char *	path		= 0;
	char *	prefix		= 0;
	char *	suffix		= 0;
	int	verbose=0;
	int 	metaheader 	= 0;
	char *	outsyntaxUID	= DC3DefaultTransferSyntax;

	GetOpt getopt(argc,argv,"d:mo:r[2]s:v");
	int c;
	int err=0;
	DC3ListofAttributes deletelist;
	DC3ListofAttributes replacelist;
	while ((c=getopt()) != -1 && !err) {
		switch (c) {
			case 'v':	// Verbose
				verbose=1;
				break;
			case 'd':	// Delete attribute keyword
				{
					char *keyword=getopt.optarg;
					char *values="";
					DC3Attribute *attr;
					DC3AttributeGet(keyword,values,
						&attr,&err);
					if (!attr || err) {
						cerr << "Couldn't handle <"
						     << keyword
						     << "> attribute\n"
						     << flush;
					}
					else {
						deletelist.insert(attr);
					}
				}
				break;
			case 'r':	// Replace attribute keyword & value
				{
					char *keyword=getopt.optarg;
					char *values=argv[getopt.optind];
					DC3Attribute *attr;
					DC3AttributeGet(keyword,values,
						&attr,&err);
					if (!attr || err) {
						cerr << "Couldn't handle <"
						     << keyword
						     << "> attribute\n"
						     << flush;
					}
					else {
						replacelist.insert(attr);
					}
				}
				break;
			case 'm':	// Generate metaheader
				metaheader=1;
				break;
			case 'o':	// Output transfer syntax
				outsyntaxUID=getopt.optarg;
				break;
			case 's':	// Stamp with specified date/time
				stamp=getopt.optarg;
				break;
			case '?':
			default:
				err=1;
		}
	}

	if (getopt.optind+4 == argc) {
		path   =argv[getopt.optind];
		series =atoi(argv[getopt.optind+1]);
		prefix =argv[getopt.optind+2];
		suffix =argv[getopt.optind+3];
	}
	else {
		cerr << "Must specify path,series,prefix & suffix\n";
		err=1;
	}

	if (err) {
		cerr << "Usage: " << argv[0] << " [-m] [-o syntax] "
		     << "[-d key] [-r key value] [-s stamp] [-v] "
		     << "inpath series outprefix outsuffix\n"
		     << flush;
		return 1;
	}

	char *nmrname;
	char *recname;
	char *drcname;

	{
		ostrstream ostr;
		ostr << path << "/" << setfill('0') << setw(4) << series << ".NMR" <<ends;
		nmrname=ostr.str();
	}
	{
		ostrstream ostr;
		ostr << path << "/" << setfill('0') << setw(4) << series << ".REC" <<ends;
		recname=ostr.str();
	}
	{
		ostrstream ostr;
		ostr << path << "/PATIENT.DRC" <<ends;
		drcname=ostr.str();
	}

	DC3TransferSyntax outtest(outsyntaxUID);
	if (!outtest.isValid()) {
		cerr << "Output Transfer Syntax UID <"
		     << outsyntaxUID << "> is invalid\n";
		return 1;
	}

	if (verbose) {
		cerr << "************** Deletions ************\n" << flush;
		deletelist.first();
		while (deletelist.ismore()) {
			DC3Attribute *attr = deletelist.value();
			attr->put(cerr);
			deletelist.next();
		}
		cerr << "************** Replacements ************\n" << flush;
		replacelist.first();
		while (replacelist.ismore()) {
			DC3Attribute *attr = replacelist.value();
			attr->put(cerr);
			replacelist.next();
		}
	}

	if (verbose) cerr << "Parsing NMR file\n" << flush;

	extern FILE *yyin;
	if ((yyin=fopen(nmrname,"r")) == 0) {
		cerr << "Can't open " << nmrname << " for read\n";
		return 1;
	}

	ExtractedNMRParameters *e = getExtractedNMRParameters();

	fclose(yyin);

	InitialString *is = 0;
	FixedFields *ff = 0;
	if (getExtractedPatientDrcParameters(drcname,is,ff) < 0 || !is || !ff)
		return 1;

	DC3ListofAttributes parameterlist;

	{
		// PatientName = ""

		parameterlist.insert(
			MakePersonNameFromString(
				DC3TagFromName(PatientName),
				extract(ff->PatientName,FixedFieldsPatientNameLength)
			)
		);

		// PatientID = ""

		parameterlist.insert(
			MakeLongStringFromString(
				DC3TagFromName(PatientID),
				extract(is->u.split.PatientId,InitialStringPatientIdLength)
			)
		);

		// ReferringPhysicianName = ""

		parameterlist.insert(
			MakePersonNameEmpty(
				DC3TagFromName(ReferringPhysicianName)
			)
		);

		// InstitutionName = ""

		parameterlist.insert(
			MakeLongStringEmpty(
				DC3TagFromName(InstitutionName)
			)
		);

		// AccessionNumber = ""

		parameterlist.insert(
			MakeShortStringEmpty(
				DC3TagFromName(AccessionNumber)
			)
		);

		// SeriesNumber = ""

		parameterlist.insert(
			MakeIntegerStringFromUnsigned(
				DC3TagFromName(SeriesNumber),
				series
			)
		);

		// StudyID = ""

		parameterlist.insert(
			MakeShortStringEmpty(
				DC3TagFromName(StudyID)
			)
		);

		// PatientSex = ""

		parameterlist.insert( 
			MakeCodeStringEmpty(
				DC3TagFromName(PatientSex)
			)
		);

		// SOPClassUID

		parameterlist.insert(
			MakeUIStringFromString(
				DC3TagFromName(SOPClassUID),
				DC3MRImageStorageSOPClass
			)
		);

		// StudyInstanceUID

		parameterlist.insert(
			MakeUIStringFromString(
				DC3TagFromName(StudyInstanceUID),
				DC3GeneratedStudyInstanceUID(stamp,study).GetUID()
			)
		);

		// SeriesInstanceUID

		parameterlist.insert(
			MakeUIStringFromString(
				DC3TagFromName(SeriesInstanceUID),
				DC3GeneratedSeriesInstanceUID(stamp,study,series).GetUID()
			)
		);

		// FrameOfReferenceUID

		parameterlist.insert(
			MakeUIStringFromString(
				DC3TagFromName(FrameOfReferenceUID),
				DC3GeneratedFrameOfReferenceUID(stamp,study,series).GetUID()
			)
		);

		// PositionReferenceIndicator = ""

		parameterlist.insert(
			MakeLongStringEmpty(DC3TagFromName(PositionReferenceIndicator))
		);

		// PixelSpacing

		parameterlist.insert(
			MakeDecimalStringFromTwoDouble(
				DC3TagFromName(PixelSpacing),
				atof(e->fov[0])/atoi(e->resolution[0]),
				atof(e->fov[1])/atoi(e->resolution[0])
			)
		);

		// ReconstructionDiameter

		parameterlist.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(ReconstructionDiameter),
				(atof(e->fov[0])>atof(e->fov[1]))
					? atof(e->fov[0])
					: atof(e->fov[1])
			)
		);

		// SliceThickness

		parameterlist.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(SliceThickness),
				atof(e->thickness)
			)
		);

		// SpacingBetweenSlices

		parameterlist.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(SpacingBetweenSlices),
				atof(e->thickness)*(atof(e->slicefactor)-1.0)
			)
		);

		// PixelBandwidth

		parameterlist.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(PixelBandwidth),
				atof(e->bwperpixel)
			)
		);

		// NumberOfAverages

		parameterlist.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(NumberOfAverages),
				atof(e->measurements)
			)
		);

		// ImagingFrequency

		parameterlist.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(ImagingFrequency),
				atof(e->freqsynth)
			)
		);

		// FlipAngle

		parameterlist.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(FlipAngle),
				atof(e->flipangle)
			)
		);

		// RepetitionTime

		parameterlist.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(RepetitionTime),
				atof((e->sequence && !strcmp(e->sequence,"IR")) ? e->trir : e->trse)
			)
		);

		// InversionTime

		if (e->sequence && !strcmp(e->sequence,"IR"))
			parameterlist.insert(
				MakeDecimalStringFromDouble(
					DC3TagFromName(InversionTime),
					atof(e->ti)
				)
			);

		// ImagedNucleus

		parameterlist.insert( 
			MakeShortStringFromString(
				DC3TagFromName(ImagedNucleus),
				e->nucleus
			)
		);

		// ReceivingCoil

		parameterlist.insert( 
			MakeShortStringFromString(
				DC3TagFromName(ReceivingCoil),
				e->coilrx
			)
		);

		// TransmittingCoil

		parameterlist.insert( 
			MakeShortStringFromString(
				DC3TagFromName(TransmittingCoil),
				e->coiltx
			)
		);

		// Modality = MR

		parameterlist.insert(
			MakeCodeStringFromString(
				DC3TagFromName(Modality),
				"MR"
			)
		);

		// ManufacturerModelName = GYROSCAN

		parameterlist.insert(
			MakeLongStringFromString(
				DC3TagFromName(ManufacturerModelName),
				"GYROSCAN"
			)
		);

		// Manufacturer = PHILIPS

		parameterlist.insert(
			MakeLongStringFromString(
				DC3TagFromName(Manufacturer),
				"PHILIPS"
			)
		);

		// PixelPaddingValue = 0

		parameterlist.insert(
			MakeUnsignedShortFromUint16(
				DC3TagFromName(PixelPaddingValue),
				0
			)
		);

		// BitsAllocated = 16

		parameterlist.insert(
			MakeUnsignedShortFromUint16(
				DC3TagFromName(BitsAllocated),
				16
			)
		);

		// BitsStored = 12

		parameterlist.insert(
			MakeUnsignedShortFromUint16(
				DC3TagFromName(BitsStored),
				12
			)
		);

		// HighBit = 11

		parameterlist.insert(
			MakeUnsignedShortFromUint16(
				DC3TagFromName(HighBit),
				11
			)
		);

		// PixelRepresentation = 0 (unsigned)

		parameterlist.insert(
			MakeUnsignedShortFromUint16(
				DC3TagFromName(PixelRepresentation),
				0
			)
		);

		// SamplesPerPixel = 1

		parameterlist.insert(
			MakeUnsignedShortFromUint16(
				DC3TagFromName(SamplesPerPixel),
				1
			)
		);

		// PhotometricInterpretation = MONOCHROME2

		parameterlist.insert(
			MakeCodeStringFromString(
				DC3TagFromName(PhotometricInterpretation),
				"MONOCHROME2"
			)
		);

		// Rows
		// Columns

		parameterlist.insert(
			MakeUnsignedShortFromUint16(
				DC3TagFromName(Rows),
				atoi(e->resolution[0])
			)
		);
		parameterlist.insert(
			MakeUnsignedShortFromUint16(
				DC3TagFromName(Columns),
				atoi(e->resolution[0])
			)
		);

		// ImageType

		parameterlist.insert(
			MakeCodeStringFromThreeStringValues(
				DC3TagFromName(ImageType),
				"ORIGINAL",
				"PRIMARY",
				"MPR"
			)
		);
	}
	{
		// ScanningSequence
		// SequenceVariant
		// ScanOptions

		DC3CodeString *ScanningSequence =
			new DC3CodeString(DC3TagFromName(ScanningSequence),0);

		DC3CodeString *SequenceVariant =
			new DC3CodeString(DC3TagFromName(SequenceVariant),0);

		DC3CodeString *ScanOptions =
			new DC3CodeString(DC3TagFromName(ScanOptions),0);
		
		if (e->sequence) {
			if (strcmp(e->sequence,"SE") == 0) {
				ScanningSequence->add("SE");
			}
			else if (strcmp(e->sequence,"IR") == 0) {
				ScanningSequence->add("IR");
			}
		}
		// else need to test for more choices :(

		if (e->presatspectral && strcmp(e->presatspectral,"no") != 0)
			ScanOptions->add("FS");
		if (e->presat && strcmp(e->presat,"no") != 0)
			ScanOptions->add("SP");
		if (e->flowcomp && strcmp(e->flowcomp,"no") != 0)
			ScanOptions->add("FC");

		SequenceVariant->first();
		if (!SequenceVariant->ismore())
			SequenceVariant->add("NONE");

		parameterlist.insert(ScanningSequence);
		parameterlist.insert(SequenceVariant);
		parameterlist.insert(ScanOptions);
	}
	{
		// MRAcquisitionType

		// Need to know more about e->scanmode ... not good :(

		DC3CodeString *MRAcquisitionType =
			new DC3CodeString(DC3TagFromName(MRAcquisitionType),0);
		
		if (e->scanmode && strcmp(e->scanmode,"MS") == 0) {
			MRAcquisitionType->add("2D");
		}

		MRAcquisitionType->first();
		if (!MRAcquisitionType->ismore())
			MRAcquisitionType->add("3D");

		parameterlist.insert(MRAcquisitionType);
	}
	{
		// EchoTrainLength = ""

		DC3Tag tag (DC3EchoTrainLength_GROUP,
			    DC3EchoTrainLength_ELEMENT);

		DC3IntegerString *attribute =
			MakeIntegerStringEmpty(tag);

		parameterlist.insert(attribute);
	}
	{
		// PatientBirthDate

		unsigned dobdd;
		unsigned dobmm;
		unsigned dobyyyy;

		getDrcDate(ff->Dob,FixedFieldsDobLength,dobdd,dobmm,dobyyyy);

		parameterlist.insert(
			MakeDateStringFromThreeValues(
				DC3TagFromName(PatientBirthDate),
				dobyyyy,
				dobmm,
				dobdd
			)
		);
	}
#ifdef CRAP
	{
		// ImageDate

		DC3TagFromName(ImageDate,
		    DC3ImageDate);

		DC3DateString *attribute =
			MakeDateStringFromThreeValues(tag,
				VISION_ImageDateYYYY.toUS(),
				VISION_ImageDateMM.toUS(),
				VISION_ImageDateDD.toUS());

		parameterlist.insert(attribute);
	}
	{
		// ImageTime

		DC3TagFromName(ImageTime,
			    DC3ImageTime);

		DC3TimeString *attribute =
			MakeTimeStringFromThreeValues(tag,
				VISION_ImageTimeHH.toUS(),
				VISION_ImageTimeMM.toUS(),
				VISION_ImageTimeSS.toUS());

		parameterlist.insert(attribute);
	}
	{
		// StudyDate

		DC3TagFromName(StudyDate,
		    DC3StudyDate);

		DC3DateString *attribute =
			MakeDateStringFromThreeValues(tag,
				VISION_RegistrationDateYYYY.toUS(),
				VISION_RegistrationDateMM.toUS(),
				VISION_RegistrationDateDD.toUS());

		parameterlist.insert(attribute);
	}
	{
		// StudyTime

		DC3TagFromName(StudyTime,
			    DC3StudyTime);

		DC3TimeString *attribute =
			MakeTimeStringFromThreeValues(tag,
				VISION_RegistrationTimeHH.toUS(),
				VISION_RegistrationTimeMM.toUS(),
				VISION_RegistrationTimeSS.toUS());

		parameterlist.insert(attribute);
	}
	{
		// SeriesDate

		DC3TagFromName(SeriesDate,
		    DC3SeriesDate);

		DC3DateString *attribute =
			MakeDateStringFromThreeValues(tag,
				VISION_SiemensStudyDateYYYY.toUS(),
				VISION_SiemensStudyDateMM.toUS(),
				VISION_SiemensStudyDateDD.toUS());

		parameterlist.insert(attribute);
	}
	{
		// SeriesTime

		DC3TagFromName(SeriesTime,
			    DC3SeriesTime);

		DC3TimeString *attribute =
			MakeTimeStringFromThreeValues(tag,
				VISION_SiemensStudyTimeHH.toUS(),
				VISION_SiemensStudyTimeMM.toUS(),
				VISION_SiemensStudyTimeSS.toUS());

		parameterlist.insert(attribute);
	}
	{
		// AcquisitionDate

		DC3TagFromName(AcquisitionDate,
		    DC3AcquisitionDate);

		DC3DateString *attribute =
			MakeDateStringFromThreeValues(tag,
				VISION_AcquisitionDateYYYY.toUS(),
				VISION_AcquisitionDateMM.toUS(),
				VISION_AcquisitionDateDD.toUS());

		parameterlist.insert(attribute);
	}
	{
		// AcquisitionTime

		DC3TagFromName(AcquisitionTime,
			    DC3AcquisitionTime);

		DC3TimeString *attribute =
			MakeTimeStringFromThreeValues(tag,
				VISION_AcquisitionTimeHH.toUS(),
				VISION_AcquisitionTimeMM.toUS(),
				VISION_AcquisitionTimeSS.toUS());

		parameterlist.insert(attribute);
	}
#endif
	// NB. assume all required parameter strings present ... :(

	// this assumes
	// - the larger resolution is first
	// - it matches display resolution ... :(

	width=atoi(e->resolution[0]);
	height=atoi(e->resolution[0]);
	unsigned nslices=atoi(e->slicesms);	// hmmm ... :(
	unsigned nechoes=atoi(e->echos);
	unsigned firstslice=nslices/2+1;
	unsigned nentries=nslices*nechoes;

	if (verbose) {
		cerr << "Reading REC file\n" << flush;
		cerr << "\twidth=\t" << width << "\n" << flush;
		cerr << "\theight=\t" << height << "\n" << flush;
		cerr << "\tnslices=\t" << nslices << "\n" << flush;
		cerr << "\tnechoes=\t" << nechoes << "\n" << flush;
		cerr << "\tfirstslice=\t" << firstslice << "\n" << flush;
		cerr << "\tnentries=\t" << nentries << "\n" << flush;
	}

	instream = new ifstream(recname,ios::in);
	if (!instream->rdbuf()->is_open()) {
		cerr << "Can't open <" << recname << "> for read\n" << flush;
		return 1;
	}

	// The 1st entry is actually the middle slice ...

	unsigned entry;
	for (entry=0; entry < nentries;++entry) {
		unsigned slice=entry/nechoes+firstslice;
		if (slice > nslices) slice-=nslices;
		unsigned echo=entry%nechoes+1;

		DC3ListofAttributes list;
		list.Replace(parameterlist);

		// SOPInstanceUID

		list.insert(
			MakeUIStringFromString(
				DC3TagFromName(SOPInstanceUID),
				DC3GeneratedSOPInstanceUID(stamp,study,series,slice).GetUID()
			)
		);

		// ImageNumber

		list.insert(
			MakeIntegerStringFromUnsigned(
				DC3TagFromName(ImageNumber),
				slice
			)
		);

		// EchoNumber

		list.insert(
			MakeIntegerStringFromUnsigned(
				DC3TagFromName(EchoNumber),
				echo
			)
		);

		// EchoTime

		list.insert(
			MakeDecimalStringFromDouble(
				DC3TagFromName(EchoTime),
				atof((echo==1)?e->te1:e->te2)
			)
		);

		// PixelData

		list.insert( 
			new DC3OtherWordString(
				DC3TagFromName(PixelData),
				width*height*2,
				reads15linestart,
				reads15line,
				reads15linedone
			)
		);

		list.Delete(deletelist);
		list.Replace(replacelist);
		list.Fix(outsyntaxUID,metaheader);

		ostrstream ostr;
		if (prefix && *prefix) ostr << prefix << ".";
		ostr << setw(1) << echo
		     << "." << setfill('0') << setw(2) << slice;
		if (suffix && *suffix) ostr << "." << suffix;
		ostr << ends;
		char *outname=ostr.str();
		if (verbose) cerr << "Writing " << outname << "\n" << flush;
		ofstream outstr(outname);

		DC3ofstream outstream(outname,outsyntaxUID);
		if (!outstream.is_open()) {
			cerr << "Can't open <" << outname
			     << "> for write\n" << flush;
			return 1;
		}

		list.first();
		while (list.ismore()) {
			DC3Attribute *attr = list.value();
			if (verbose) {
				attr->put(cerr);
			}
			attr->write(outstream);
			list.next();
		}

		outstream.close();
	}

	instream->close();

	return 0;
}

