static char *CopyrightIdentifier(void) { return "@(#)sgntodc.cc 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved."; }

#ifdef __GNUG__
#include <builtin.h>
#endif
#include <math.h>
#include <string.h>

#include <strstream.h>
#include <iomanip.h>

#include "tlist.h"
#include "tfstream.h"
#include "tddictc.h"

#include "dgtypes.h"

#include "GetOpt.h"
#include "MainCmd.h"

static ifstream	*instream;

static Uint16 	resolution;
static Uint16 	datablock;
static Uint16 	*linebuffer;

void
read_header(
	ifstream& instream,		// where to read from
	DC3ListofAttributes& list,	// list to add new attributes to
	const char *stamp,		// substitute date/time stamp for UIDs
	Uint16& resolution,		// Returned - image resolution
	Uint16& datablock		// Returned - block offset to image
);

static Uint16
readUint16(ifstream *instream)
{
	Uint16 u;
	unsigned char buffer[2];
	instream->read(buffer,2);
	// DataGeneral is a Big Endian machine
	u =  (Uint16)buffer[0];
	u <<= 8;
	u |= (Uint16)buffer[1];
	return u;
}

static streampos
seekblock(istream *instream,unsigned block)
{
	const unsigned offset = block*256*2;	// 256 word blocks,16 bit words
	return instream->rdbuf()->seekoff
			(offset,ios::beg,ios::in|ios::out);
}

static int
readsignalinestart(void)
{
	// instream is already open

	instream->clear();	// Clear EOF state to allow seeking/reading

	// Position at start of image pixel data ...

	if (seekblock(instream,datablock) == streampos(EOF)) {
		cerr << "Image seek failed\n" << flush;
		return -1;
	}

	// Initialize decompression parameters ...

	linebuffer=new Uint16[resolution];
	if (!linebuffer) {
		cerr << "Couldn't allocate linebuffer\n" << flush;
		return -1;
	}

	return 0;
}

static int
readsignalinedone(void)
{
	// leave instream open - closed in main routine

	if (linebuffer)	delete [] linebuffer;

	return 0;
}

static int
readsignaline(Uint16 **buffer)
{
	Uint16 *ptr = linebuffer;

	unsigned col;
	for (col=0; col<resolution; ++col) {
		Uint16 u=readUint16(instream);
		if (!instream->good()) break;
		*ptr++=u;
	}

	if (!instream->good()) {
		cerr << "Line read failed\n" << flush;
		return -1;
	}

	*buffer=linebuffer;		// return pointer to data read
	return resolution;		// return length of data (1 line)
}

int
main(int argc,char *argv[])
{
	CCOMMAND(argc,argv);

	char *	stamp		= 0;
	char *	inname		= 0;
	char *	outname		= 0;
	int 	verbose		= 0;
	int 	metaheader 	= 0;
	char *	outsyntaxUID	= DC3DefaultTransferSyntax;

	GetOpt getopt(argc,argv,"d:mo:r[2]s:v");
	int c;
	int err=0;
	DC3ListofAttributes deletelist;
	DC3ListofAttributes replacelist;
	while ((c=getopt()) != -1 && !err) {
		switch (c) {
			case 'v':	// Verbose
				verbose=1;
				break;
			case 'd':	// Delete attribute keyword
				{
					char *keyword=getopt.optarg;
					char *values="";
					DC3Attribute *attr;
					DC3AttributeGet(keyword,values,
						&attr,&err);
					if (!attr || err) {
						cerr << "Couldn't handle <"
						     << keyword
						     << "> attribute\n"
						     << flush;
					}
					else {
						deletelist.insert(attr);
					}
				}
				break;
			case 'r':	// Replace attribute keyword & value
				{
					char *keyword=getopt.optarg;
					char *values=argv[getopt.optind];
					DC3Attribute *attr;
					DC3AttributeGet(keyword,values,
						&attr,&err);
					if (!attr || err) {
						cerr << "Couldn't handle <"
						     << keyword
						     << "> attribute\n"
						     << flush;
					}
					else {
						replacelist.insert(attr);
					}
				}
				break;
			case 'm':	// Generate metaheader
				metaheader=1;
				break;
			case 'o':	// Output transfer syntax
				outsyntaxUID=getopt.optarg;
				break;
			case 's':	// Stamp with specified date/time
				stamp=getopt.optarg;
				break;
			case '?':
			default:
				err=1;
		}
	}

	if (getopt.optind+2 == argc) {
		inname=argv[getopt.optind];
		outname=argv[getopt.optind+1];
	}
	else err=1;

	if (err) {
		cerr << "Usage: " << argv[0] << " [-m] [-o syntax] "
		     << "[-d key] [-r key value] [-s stamp] [-v] "
		     << "infile outfile\n"
		     << flush;
		return 1;
	}

	if (verbose) {
		cerr << "************** Deletions ************\n" << flush;
		deletelist.first();
		while (deletelist.ismore()) {
			DC3Attribute *attr = deletelist.value();
			attr->put(cerr);
			deletelist.next();
		}
		cerr << "************** Replacements ************\n" << flush;
		replacelist.first();
		while (replacelist.ismore()) {
			DC3Attribute *attr = replacelist.value();
			attr->put(cerr);
			replacelist.next();
		}
	}

	instream = new ifstream(inname,ios::in);
	if (!instream->rdbuf()->is_open()) {
		cerr << "Can't open <" << inname << "> for read\n" << flush;
		return 1;
	}

	DC3ListofAttributes list;

	if (verbose)
		cerr << "**************** Reading **************\n" << flush;

	read_header(*instream,list,stamp,resolution,datablock);

	{
		// Add a PixelData attribute ...

		DC3Tag tag (DC3PixelData_GROUP,DC3PixelData_ELEMENT);
		DC3OtherWordString *attr = 
			new DC3OtherWordString(tag,
				resolution*resolution*2,
				readsignalinestart,
				readsignaline,
				readsignalinedone);
		list.insert(attr);
	}

	if (verbose)
		cerr << "**************** Deleting **************\n" << flush;

	list.Delete(deletelist);

	if (verbose)
		cerr << "**************** Replacing **************\n" << flush;

	list.Replace(replacelist);

	if (verbose)
		cerr << "**************** Fixing **************\n" << flush;

	list.Fix(outsyntaxUID,metaheader);

	if (verbose)
		cerr << "**************** Writing **************\n" << flush;

	DC3TransferSyntax outtest(outsyntaxUID);
	if (!outtest.isValid()) {
		cerr << "Output Transfer Syntax UID <"
		     << outsyntaxUID << "> is invalid\n";
		return 1;
	}

	DC3ofstream outstream(outname,outsyntaxUID);
	if (!outstream.is_open()) {
		cerr << "Can't open <" << outname << "> for write\n" << flush;
		return 1;
	}

	list.first();
	while (list.ismore()) {
		DC3Attribute *attr = list.value();
		if (verbose) {
			attr->put(cerr);
		}
		attr->write(outstream);
		list.next();
	}

	outstream.close();

	return 0;
}

