static char *CopyrightIdentifier(void) { return "@(#)pacetopgm.cc 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved."; }

#include <fstream.h>

#include "MainCmd.h"

typedef unsigned short	Uint16;
typedef short		Int16;

static Uint16
read16(istream& instream)
{
	Uint16 u;
	unsigned char buffer[2];
	instream.read(buffer,2);
	// Pace is a Big Endian machine
	u =  (Uint16)buffer[0];
	u <<= 8;
	u |= (Uint16)buffer[1];
	return u;
}

static void
write16little(ostream& outstream,Uint16 u)
{
#ifdef	PNMBIGRAW
	unsigned char buffer[2];
	// Output is little endian
	buffer[0]=(unsigned char)(u);
	buffer[1]=(unsigned char)(u>>8);
	outstream.write(buffer,2);
#else /* PNMBIGRAW */
	outstream << u << "\n";
#endif /* PNMBIGRAW */
}

static streampos
seekword(istream& instream,unsigned offset,unsigned word)
{
	// byte offset from 0 ...
	return instream.rdbuf()->seekoff
			(offset+word,ios::beg,ios::in|ios::out);
}

static void
copypaceimage(istream& instream,ostream& outstream,
	      Uint16 width,Uint16 height)
{
// NB. the exclusive or with 0x8000 makes the signed Pace values unsigned
// which is what the PGM convention is ... just omit the ^0x8000
// everywhere if you want the data left signed.

	unsigned i;
	Int16 pixel=0;
	enum Mode { Difference, Reference } mode = Reference;
	for (i=0; i<height*width;) {
		unsigned char byte;
		instream.read(&byte,1);
		if (!instream) break;
		if (byte == 0x80) {		// Mode switch
			if (mode == Difference)
				mode=Reference;
			else
				mode=Difference;
		}
		else if (byte == 0x81) {	// Run length flag
			instream.read(&byte,1);
			if (!instream) break;
			unsigned repeat=byte;
			i+=repeat;
			while (repeat--) write16little(outstream,pixel^0x8000);
		}
		else {
			if (mode == Difference) {
				pixel+=(signed char)byte;
			}
			else {
				pixel=byte<<8;
				instream.read(&byte,1);
				if (!instream) break;
				pixel|=byte;
			}
			write16little(outstream,pixel^0x8000);
			++i;
		}
	}
	if (!instream) cerr << "Premature EOF byte " << i << "\n" << flush;
}

int
main(int argc,char *argv[])
{
	CCOMMAND(argc,argv);

	if (argc != 1) {
		cerr << "Usage: " << argv[0] << "< infile > outfile\n" << flush;
		return 1;
	}

	// default for some environments is unbuffered which is slow++
	// the following improves performance 25 fold with libg++ ...

	const unsigned bufsize=4096;
	streambuf *in  = cin.rdbuf();
	streambuf *out = cout.rdbuf();
	static char inbuf[bufsize];
	static char outbuf[bufsize];
	in->setbuf(inbuf,sizeof(inbuf));
	out->setbuf(outbuf,sizeof(outbuf));

	unsigned imageheaderblock=768;
	unsigned datablock=1792;

	(void)seekword(cin,imageheaderblock,416);
	unsigned width=read16(cin);

	(void)seekword(cin,imageheaderblock,418);
	unsigned height=read16(cin);

#ifdef	PNMBIGRAW
	cout << "P5\n" << width << " " << height << "\n"
		    << 4095 << "\n";
#else /* PNMBIGRAW */
	cout << "P2\n" << width << " " << height << "\n"
		    << 4095 << "\n";
#endif /* PNMBIGRAW */

	// Decompress and output the image data ...

	if (seekword(cin,datablock,0) == streampos(EOF)) {
		cerr << "Image seek failed\n" << flush;
		return 1;
	}
	else {
		copypaceimage(cin,cout,width,height);
		if (!cin) {
			cerr
			  << "Image read failed\n"
			  << flush;
			return 1;
		}
		if (!cout.good()) {
			cerr
			  << "Image write failed\n"
			  << flush;
			return 1;
		}
	}

	return 0;
}

