#  condition.awk 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved.

# create a C++ header or method file to generate conditions
# according to a description in the input template file

# can set these values on the command line:
#
#	role	  - either "header" or "method" (defaults to "header")

NR==1	{
	print "// Automatically generated from template - EDITS WILL BE LOST"
	print ""
	print "// Generated by condition.awk with options " role
	print ""

	if (role == "") role   = "header"

	# Never and Always need to be callable as functions ...

	if (role == "header") {
		print "#ifndef _DC3tcond_HEADER_"
		print "#define _DC3tcond_HEADER_"
		print ""
		print "int DC3Condition_Never(DC3ListofAttributes *list);"
		print "int DC3Condition_Always(DC3ListofAttributes *list);"
	}
	else if (role == "method") {
		print "#include \"tcondt.h\""
		print "#include \"tcond.h\""
		print ""
		print "int DC3Condition_Never(DC3ListofAttributes *list)  { (void)list; return 0; }"
		print "int DC3Condition_Always(DC3ListofAttributes *list) { (void)list; return 1; }"
	}
	else {
		print "Error - role " role " invalid" >"/dev/tty"
		exit 1
	}

	condition=""
	}

/^[ 	]*Condition=/ {

	condition=""
	if (match($0,"Condition=\"[^\"]*\""))
		condition=substr($0,RSTART+length("Condition=\""),
			RLENGTH-length("Condition=\"")-1);

	if (role == "header") {
		print "int	DC3Condition_" condition "(DC3ListofAttributes *list);"
	}
	else if (role == "method") {
		print "int"
		print "DC3Condition_" condition "(DC3ListofAttributes *list)"
		print "{"
		print "\tint condition=0;\t// default is false"
		print "\t(void)list;"
		print ""
	}

	}

/^[ 	]*ConditionEnd/ {

	if (role == "method") {
		print ""
		print "\treturn condition;"
		print "}"
		print ""
	}
	condition=""

	}

/^[ 	]*Element/ {

	element=""
	if (match($0,"Element=\"[^\"]*\""))
		element=substr($0,RSTART+length("Element=\""),
			RLENGTH-length("Element=\"")-1);

	elementpresent=0
	elementpresentmask=""
	if (match($0,"ElementPresent=\"[^\"]*\"")) {
		elementpresent=1;
		elementpresentmask=substr($0,RSTART+length("ElementPresent=\""),
			RLENGTH-length("ElementPresent=\"")-1);
	}

	andelementpresent=0
	andelementpresentmask=""
	if (match($0,"AndElementPresent=\"[^\"]*\"")) {
		andelementpresent=1;
		andelementpresentmask=substr($0,RSTART+length("AndElementPresent=\""),
			RLENGTH-length("AndElementPresent=\"")-1);
	}

	notelementpresent=0
	notelementpresentmask=""
	if (match($0,"NotElementPresent=\"[^\"]*\"")) {
		notelementpresent=1;
		notelementpresentmask=substr($0,RSTART+length("NotElementPresent=\""),
			RLENGTH-length("NotElementPresent=\"")-1);
	}

	# Should fix substring match but this will do for now ...
	if (andelementpresent == 1 || notelementpresent == 1) {
		elementpresent=0;
		elementpresentmask=""
	}

	grouppresent=0;
	grouppresentmask=""
	if (match($0,"GroupPresent=\"[^\"]*\"")) {
		grouppresent=1;
		grouppresentmask=substr($0,RSTART+length("GroupPresent=\""),
			RLENGTH-length("GroupPresent=\"")-1);
	}

	stringconstant=""
	if (match($0,"StringConstant=\"[^\"]*\""))
		stringconstant=substr($0,RSTART+length("StringConstant=\""),
			RLENGTH-length("StringConstant=\"")-1);

	stringvalue=""
	if (match($0,"StringValue=\"[^\"]*\""))
		stringvalue=substr($0,RSTART+length("StringValue=\""),
			RLENGTH-length("StringValue=\"")-1);

	binaryvalue=""
	if (match($0,"BinaryValue=\"[^\"]*\""))
		binaryvalue=substr($0,RSTART+length("BinaryValue=\""),
			RLENGTH-length("BinaryValue=\"")-1);

	if (role == "method") {
		if (elementpresent) {
			if (length(elementpresentmask) > 0) {
				print "\tcondition|=ElementPresentMasked(list,DC3TagFromName(" element ")," elementpresentmask ");"
			}
			else {
				print "\tcondition|=ElementPresent(list,DC3TagFromName(" element "));"
			}
		}
		if (andelementpresent) {
			if (length(andelementpresentmask) > 0) {
				print "\tcondition&=ElementPresentMasked(list,DC3TagFromName(" element ")," andelementpresentmask ");"
			}
			else {
				print "\tcondition&=ElementPresent(list,DC3TagFromName(" element "));"
			}
		}
		if (notelementpresent) {
			if (length(notelementpresentmask) > 0) {
				print "\tcondition|=!ElementPresentMasked(list,DC3TagFromName(" element ")," notelementpresentmask ");"
			}
			else {
				print "\tcondition|=!ElementPresent(list,DC3TagFromName(" element "));"
			}
		}
		if (grouppresent) {
			if (length(grouppresentmask) > 0) {
				print "\tcondition|=GroupPresentMasked(list,DC3TagFromName(" element ")," grouppresentmask ");"
			}
			else {
				print "\tcondition|=GroupPresent(list,DC3TagFromName(" element "));"
			}
		}
		if (length(stringconstant) > 0) {
			print "\tcondition|=StringValueMatch(list,DC3TagFromName(" element "),DC3" stringconstant ");"
		}
		if (length(stringvalue) > 0) {
			print "\tcondition|=StringValueMatch(list,DC3TagFromName(" element "),\"" stringvalue "\");"
		}
		if (length(binaryvalue) > 0) {
			print "\t{"
			print "\t\tXint value;"
			print "\t\tif (BinaryValueGet(list,DC3TagFromName(" element "),value))"
			print "\t\t\tcondition|=(value " binaryvalue ");"
			print "\t}"
		}
	}

	}

END {
	if (role == "header") {
		print ""
		print "#endif /* _DC3tcond_HEADER_ */"
	}
}

