#  tpltohdr.awk 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved.

# create a C++ header file to read keywords & make dicom attributes
# according to a description in the input template file

# goes to great effort to make sure things are read
#	- in order
#	- once
# so that theoretically the C++ routine could read from standard input
# (assuming of course that the template file is in ascending block
# and offset order)

# can set these values on the command line:
#
#	prefix	  - the prefix to insert before keywords (defaults to "PREFIX")
#	role	  - either "dump" or "convert" (defaults to "convert")
#	splittype - to generate multiple headers rather than monolithic one
#		  - "declare"   - output global keyword declarations
#		  - "define"    - output global keyword definitions
#		  - "functions" - output lots of little function definitions
#		  - "calls"     - output series of calls to little functions

NR==1	{
	if (prefix == "") prefix = "PREFIX"
	if (role   == "") role   = "convert"
	if (splittype == "")
		single = "yes"
	else
		single = "no"

	print "// Automatically generated from template - EDITS WILL BE LOST"
	print ""
	print "// Generated by tpltohdr.awk with options " \
			prefix " " role " " splittype
	print ""
	}

/^block/{
	block=""
	if (match($0,"block=\"[^\"]*\""))
		block=substr($0,RSTART+length("block=\""),
			RLENGTH-length("block=\"")-1);
	offset=""
	if (match($0,"offset=\"[^\"]*\""))
		offset=substr($0,RSTART+length("offset=\""),
			RLENGTH-length("offset=\"")-1);
	intype=""
	if (match($0,"intype=\"[^\"]*\""))
		intype=substr($0,RSTART+length("intype=\""),
			RLENGTH-length("intype=\"")-1);
	inlength=""
	if (match($0,"inlength=\"[^\"]*\""))
		inlength=substr($0,RSTART+length("inlength=\""),
			RLENGTH-length("inlength=\"")-1);
	keyword=""
	if (match($0,"keyword=\"[^\"]*\""))
		keyword=substr($0,RSTART+length("keyword=\""),
			RLENGTH-length("keyword=\"")-1);
	name=""
	if (match($0,"name=\"[^\"]*\""))
		name=substr($0,RSTART+length("name=\""),
			RLENGTH-length("name=\"")-1);
	dicomtype=""
	if (match($0,"dicomtype=\"[^\"]*\""))
		dicomtype=substr($0,RSTART+length("dicomtype=\""),
			RLENGTH-length("dicomtype=\"")-1);
	dicomtag=""
	if (match($0,"dicomtag=\"[^\"]*\""))
		dicomtag=substr($0,RSTART+length("dicomtag=\""),
			RLENGTH-length("dicomtag=\"")-1);
	enum=""
	if (match($0,"enum=\"[^\"]*\""))
		enum=substr($0,RSTART+length("enum=\""),
			RLENGTH-length("enum=\"")-1);
	bitmap=""
	if (match($0,"bitmap=\"[^\"]*\""))
		bitmap=substr($0,RSTART+length("bitmap=\""),
			RLENGTH-length("bitmap=\"")-1);
	comment=""
	if (match($0,"#.*$"))
		comment=substr($0,RSTART+length("#"),
			RLENGTH-length("#"));

	ok="yes";

	if (block == "" || block == "?" ) {
		print "Line " FNR ": error - no block" >"/dev/tty"
		ok="no";
	}
	if (offset == "" || offset == "?" ) {
		print "Line " FNR ": error - no offset" >"/dev/tty"
		ok="no";
	}
	if (intype == "" || intype == "?" ) {
		print "Line " FNR ": error - no intype" >"/dev/tty"
		ok="no";
	}
	if (inlength == "" || inlength == "?" ) {
		print "Line " FNR ": error - no inlength" >"/dev/tty"
		ok="no";
	}

	if (role == "convert") {
		if (dicomtype != "" && dicomtype != "?") {
			if      (dicomtype == "LO") { dicomtypedesc="LongString"    ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "PN") { dicomtypedesc="PersonName"    ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "SH") { dicomtypedesc="ShortString"   ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "IS") { dicomtypedesc="IntegerString" ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "DS") { dicomtypedesc="DecimalString" ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "DA") { dicomtypedesc="DateString"    ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "TM") { dicomtypedesc="TimeString"    ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "CS") { dicomtypedesc="CodeString"    ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "LT") { dicomtypedesc="LongText"      ; outtype="char *"; args=",strlen(out),out"; }
			else if (dicomtype == "US") { dicomtypedesc="UnsignedShort" ; outtype="Uint16"; args=",2,&out"; }
			else if (dicomtype == "XS") { dicomtypedesc="UnspecifiedShort" ; outtype="Uint16"; args=",2,&out"; }
			else {
				print "Line " FNR ": error in dicomtype - bad type <" \
					dicomtype ">" >"/dev/tty"
				ok="no";
			}
			putdicom="yes"
		}
		else {
			putdicom="no"
		}
	}

	if (role == "dump") {
		if      (intype == "VAX_char" )      { outtype="char *"; outclass="SH"; }
		else if (intype == "VAX_u_byte" )    { outtype="Uint16"; outclass="US"; }
		else if (intype == "VAX_byte" )      { outtype="Int16";  outclass="SS"; }
		else if (intype == "VAX_u_short" )   { outtype="Uint16"; outclass="US"; }
		else if (intype == "VAX_short" )     { outtype="Int16";  outclass="SS"; }
		else if (intype == "VAX_int" )       { outtype="Int32";  outclass="SL"; }
		else if (intype == "VAX_u_int" )     { outtype="Uint32"; outclass="UL"; }
		else if (intype == "VAX_float_f" )   { outtype="char *"; outclass="DS"; }
		else if (intype == "SUN3_char" )     { outtype="char *"; outclass="SH"; }
		else if (intype == "SUN3_u_short" )  { outtype="Uint16"; outclass="US"; }
		else if (intype == "SUN3_short" )    { outtype="Int16";  outclass="SS"; }
		else if (intype == "SUN3_int" )      { outtype="Int32";  outclass="SL"; }
		else if (intype == "SUN3_u_int" )    { outtype="Uint32"; outclass="UL"; }
		else if (intype == "SUN3_float" )    { outtype="char *"; outclass="DS"; }
		else if (intype == "SUN3_double" )   { outtype="char *"; outclass="DS"; }
		else if (intype == "SUN3_datetime")  { outtype="char *"; outclass="DT"; }
		else if (intype == "DGA" )           { outtype="char *"; outclass="SH"; }
		else if (intype == "DGI" )           { outtype="Int16"; outclass="US"; }
		else if (intype == "DGI4" )          { outtype="Int32"; outclass="UL"; }
		else if (intype == "DGR" )           { outtype="char *"; outclass="DS"; }
		else if (intype == "DGA2" )          { outtype="char *"; outclass="SH"; }
		else if (intype == "DGDD" )          { outtype="char *"; outclass="SH"; }
		else if (intype == "DGTT" )          { outtype="char *"; outclass="SH"; }
		else if (intype == "SYTEC_byte" )    { outtype="Uint16"; outclass="US"; }
		else if (intype == "SYTEC_string" )  { outtype="char *"; outclass="SH"; }
		else if (intype == "SYTEC_int16" )   { outtype="Int16";  outclass="SS"; }
		else if (intype == "SYTEC_int32" )   { outtype="Int32";  outclass="SL"; }
		else if (intype == "SYTEC_u_int16" ) { outtype="Uint16"; outclass="US"; }
		else if (intype == "SYTEC_u_int32" ) { outtype="Uint32"; outclass="UL"; }
		else if (intype == "SYTEC_float32" ) { outtype="char *"; outclass="DS"; }
		else if (intype == "SYTEC_datetime") { outtype="char *"; outclass="DT"; }
		else if (intype == "PACE_byte" )     { outtype="Uint16"; outclass="US"; }
		else if (intype == "PACE_string" )   { outtype="char *"; outclass="SH"; }
		else if (intype == "PACE_int16" )    { outtype="Int16";  outclass="SS"; }
		else if (intype == "PACE_int32" )    { outtype="Int32";  outclass="SL"; }
		else if (intype == "PACE_u_int16" )  { outtype="Uint16"; outclass="US"; }
		else if (intype == "PACE_u_int32" )  { outtype="Uint32"; outclass="UL"; }
		else if (intype == "PACE_float32" )  { outtype="char *"; outclass="DS"; }
		else if (intype == "PACE_int_date")  { outtype="char *"; outclass="DA"; }
		else if (intype == "PACE_int_time")  { outtype="char *"; outclass="TM"; }
		else if (intype == "PACE_string_date")  { outtype="char *"; outclass="DA"; }
		else if (intype == "PACE_string_time")  { outtype="char *"; outclass="TM"; }
		else {
			print "Line " FNR ": error in intype - bad type <" \
				intype ">" >"/dev/tty"
			ok="no";
		}
	}

	if (keyword != "" && keyword != "?") {
		usename=prefix "_" keyword
		visible="yes"
	}
	else {
		if (name != "" && name != "?") {
			usename=prefix "_" name
		}
		else {
			usename="temp"
		}
		visible="no"
	}
	useblock=prefix "_" block

	if (ok == "yes") {
		if (visible == "yes" || role == "dump" || putdicom == "yes") {
			if (single == "no" && splittype == "functions") {
				print "void"
				if (role == "convert") print prefix "_do_" block "_" offset \
					"(ifstream& instream,DC3ListofAttributes& list)"
				if (role == "dump") print prefix "_do_" block "_" offset \
					"(ifstream& instream)"
				print "{"
				if (role == "convert")
					print "\t(void)list;	// Shuts up warnings if not used"
				print ""
			}
			print "\t// " comment
			if (visible == "yes") {
				if (single == "yes" || splittype == "define") {
					print ""
					print "\tconst unsigned\t" \
						usename "_length = " inlength ";"
					print "\t" intype "\t" usename " (" \
						usename "_length);"
				}
				if (single == "no" && splittype == "declare") {
					print ""
					print "\textern const unsigned\t" usename "_length;"
					print "\textern " intype "\t" usename ";"
				}
			}
			print ""
			if (single == "yes" || splittype == "functions") {
				print "\tif (" useblock "_lng) {"
				if (visible == "no") {
					print "\t\tconst unsigned\t" \
						usename "_length = " inlength ";"
					print "\t\t" intype "\t" usename " (" \
						usename "_length);"
				}
				print "\t\tconst unsigned blkptr = " useblock "_ptr;"
				print "\t\tconst unsigned offset = " offset ";"
				print "\t\t" usename ".read(instream,blkptr,offset);"

				if (role == "convert" && putdicom == "yes") {
					print ""
					print "\t\t// DICOM - " dicomtag
					print ""
					print "\t\tDC3Tag\ttag(DC3" \
						dicomtag "_GROUP,DC3" dicomtag "_ELEMENT);"
					print "\t\t" outtype "\tout = " usename \
						".to" dicomtype "();"
					print "\t\tDC3" dicomtypedesc " *attribute ="
					print "\t\t\tnew DC3" dicomtypedesc "(tag" args ");"
					print "\t\tlist.insert(attribute);"
				}

				if (role == "dump") {
					print "\t\t" outtype "\tout = " usename ".to" outclass "();"
					if (keyword != "" && keyword != "?") keydesc=" (" keyword ")"
					else keydesc=""

					print "\t\tcerr << \"[\" << blkptr << \":\" << offset << \"]" \
						"\\t" comment keydesc "\\t <\" << out << \">\\n\";"

					while (enum != "") {
						#if (!match(enum,"[a-zA-Z\'0-9\-][a-zA-Z\'0-9]*=")) {
						if (!match(enum,"[^=]*=")) {
							print "Line " FNR \
								": error in enum - no code <" \
								enum ">" >"/dev/tty"
							break
						}
						code=substr(enum,RSTART,RLENGTH-1)
						enum=substr(enum,RSTART+RLENGTH)
						if (!match(enum,"[^,][^,]*")) {
							print "Line " FNR \
								": error in enum - no value <" \
								enum ">" >"/dev/tty"
							break
						}
						meaning=substr(enum,RSTART,RLENGTH)
						enum=substr(enum,RSTART+RLENGTH)
						if (match(code,"\'")) {
							unquotecode=substr(code,2,length(code)-2);
							print "\t\tif (strcmp(out,\"" unquotecode \
								"\")==0) cerr << \"\\t\\t " \
								code " = " meaning "\\n\";"
						}
						else {
							print "\t\tif (out == " code \
								") cerr << \"\\t\\t\\t " \
								code " = " meaning "\\n\";"
						}
						if (match(enum,"^,")) {
							enum=substr(enum,RSTART+RLENGTH)
						}
						else {
							if (length(enum) != 0) {
							print "Line " FNR \
								": error in enum - trailing garbage <" \
								enum ">" >"/dev/tty"
							}
							break;
						}
					}

					while (bitmap != "") {
						if (!match(bitmap,"[0-9][0-9]*:")) {
							print "Line " FNR \
								": error in bitmap - no bit number <" \
								bitmap ">" >"/dev/tty"
							break
						}
						bitnumber=substr(bitmap,RSTART,RLENGTH-1)
						bitmap=substr(bitmap,RSTART+RLENGTH)

						if (!match(bitmap,"[^,;][^,;]*")) {
							print "Line " FNR \
								": error in bitmap - no false value <" \
								bitmap ">" >"/dev/tty"
							break
						}
						zeromeaning=substr(bitmap,RSTART,RLENGTH)
						bitmap=substr(bitmap,RSTART+RLENGTH)

						if (match(bitmap,"^,")) {
							bitmap=substr(bitmap,RSTART+RLENGTH)
						}
						else {
							print "Line " FNR \
								": error in bitmap - no comma <" \
								bitmap ">" >"/dev/tty"
							break;
						}

						if (!match(bitmap,"[^;,][^;,]*")) {
							print "Line " FNR \
								": error in bitmap - no true value <" \
								bitmap ">" >"/dev/tty"
							break
						}
						onemeaning=substr(bitmap,RSTART,RLENGTH)
						bitmap=substr(bitmap,RSTART+RLENGTH)

						print "\t\t{"
						print "\t\t\t" outtype \
							" bitvalue=out&(1<<" bitnumber ");"
						print "\t\t\tcerr << \"\\t\\t\\t bit " bitnumber \
							" \" << (bitvalue?\"true \":\"false\") " \
							"<< \" = \" << (bitvalue?\"" onemeaning \
							"\":\"" zeromeaning "\") << \"\\n\";"
						print "\t\t}"

						if (match(bitmap,"^;")) {
							bitmap=substr(bitmap,RSTART+RLENGTH)
						}
						else {
							if (length(bitmap) != 0) {
							print "Line " FNR \
							    ": error in bitmap - trailing garbage <" \
								bitmap ">" >"/dev/tty"
							}
							break;
						}
					}
				}

				print "\t}"
				print ""
			}
			if (single == "no" && splittype == "calls") {
				if (role == "convert") {
					print "\tvoid " prefix "_do_" block "_" offset \
						"(ifstream& instream,DC3ListofAttributes& list);"
					print "\t" prefix "_do_" block "_" offset "(instream,list);"
				}
				if (role == "dump") {
					print "\tvoid " prefix "_do_" block "_" offset \
						"(ifstream& instream);"
					print "\t" prefix "_do_" block "_" offset "(instream);"
				}
				print ""
			}
			if (single == "no" && splittype == "functions") {
				print "}"
				print ""
			}
		}
	}

	}
